# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test_rpad_and_clip_empty_array():
    empty = ak.contents.emptyarray.EmptyArray()
    assert to_list(empty) == []
    assert to_list(ak._do.pad_none(empty, 5, 0)) == [None, None, None, None, None]
    assert (
        ak._do.pad_none(empty.to_typetracer(), 5, 0).form
        == ak._do.pad_none(empty, 5, 0).form
    )
    assert to_list(ak._do.pad_none(empty, 5, 0, clip=True)) == [
        None,
        None,
        None,
        None,
        None,
    ]
    assert (
        ak._do.pad_none(empty.to_typetracer(), 5, 0, clip=True).form
        == ak._do.pad_none(empty, 5, 0, clip=True).form
    )


def test_rpad_and_clip_numpy_array():
    array = ak.contents.numpyarray.NumpyArray(
        np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)
    )
    assert to_list(array) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]

    assert to_list(ak._do.pad_none(array, 5, 0, clip=True)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
        None,
        None,
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 0, clip=True).form
        == ak._do.pad_none(array, 5, 0, clip=True).form
    )

    assert to_list(ak._do.pad_none(array, 5, 1, clip=True)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14], None, None],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29], None, None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 1, clip=True).form
        == ak._do.pad_none(array, 5, 1, clip=True).form
    )

    assert to_list(ak._do.pad_none(array, 7, 2, clip=True)) == [
        [
            [0, 1, 2, 3, 4, None, None],
            [5, 6, 7, 8, 9, None, None],
            [10, 11, 12, 13, 14, None, None],
        ],
        [
            [15, 16, 17, 18, 19, None, None],
            [20, 21, 22, 23, 24, None, None],
            [25, 26, 27, 28, 29, None, None],
        ],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 7, 2, clip=True).form
        == ak._do.pad_none(array, 7, 2, clip=True).form
    )

    assert to_list(ak._do.pad_none(array, 2, 2, clip=True)) == [
        [[0, 1], [5, 6], [10, 11]],
        [[15, 16], [20, 21], [25, 26]],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 2, 2, clip=True).form
        == ak._do.pad_none(array, 2, 2, clip=True).form
    )


def test_rpad_numpy_array():
    array = ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    assert to_list(ak._do.pad_none(array, 10, 0)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        None,
        None,
        None,
        None,
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 10, 0).form
        == ak._do.pad_none(array, 10, 0).form
    )

    array = ak.contents.numpyarray.NumpyArray(
        np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])
    )
    assert to_list(ak._do.pad_none(array, 5, 0)) == [
        [1.1, 2.2, 3.3],
        [4.4, 5.5, 6.6],
        None,
        None,
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 0).form
        == ak._do.pad_none(array, 5, 0).form
    )
    assert to_list(ak._do.pad_none(array, 5, 1)) == [
        [1.1, 2.2, 3.3, None, None],
        [4.4, 5.5, 6.6, None, None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 1).form
        == ak._do.pad_none(array, 5, 1).form
    )

    array = ak.contents.numpyarray.NumpyArray(
        np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)
    )
    assert to_list(array) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]

    assert to_list(ak._do.pad_none(array, 1, 0)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]

    assert (
        ak._do.pad_none(array.to_typetracer(), 1, 0).form
        == ak._do.pad_none(array, 1, 0).form
    )
    assert to_list(ak._do.pad_none(array, 2, 0)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 2, 0).form
        == ak._do.pad_none(array, 2, 0).form
    )
    assert to_list(ak._do.pad_none(array, 3, 0)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 3, 0).form
        == ak._do.pad_none(array, 3, 0).form
    )
    assert to_list(ak._do.pad_none(array, 4, 0)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
        None,
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 4, 0).form
        == ak._do.pad_none(array, 4, 0).form
    )
    assert to_list(ak._do.pad_none(array, 5, 0)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
        None,
        None,
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 0).form
        == ak._do.pad_none(array, 5, 0).form
    )

    assert to_list(ak._do.pad_none(array, 2, 1)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]
    assert to_list(ak._do.pad_none(array, 3, 1)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 3, 1).form
        == ak._do.pad_none(array, 3, 1).form
    )
    assert to_list(ak._do.pad_none(array, 4, 1)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14], None],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29], None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 4, 1).form
        == ak._do.pad_none(array, 4, 1).form
    )
    assert to_list(ak._do.pad_none(array, 5, 1)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14], None, None],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29], None, None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 1).form
        == ak._do.pad_none(array, 5, 1).form
    )

    assert to_list(ak._do.pad_none(array, 3, 2)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 3, 2).form
        == ak._do.pad_none(array, 3, 2).form
    )

    assert to_list(ak._do.pad_none(array, 7, 2)) == [
        [
            [0, 1, 2, 3, 4, None, None],
            [5, 6, 7, 8, 9, None, None],
            [10, 11, 12, 13, 14, None, None],
        ],
        [
            [15, 16, 17, 18, 19, None, None],
            [20, 21, 22, 23, 24, None, None],
            [25, 26, 27, 28, 29, None, None],
        ],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 7, 2).form
        == ak._do.pad_none(array, 7, 2).form
    )

    assert to_list(ak._do.pad_none(array, 2, 2)) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 2, 2).form
        == ak._do.pad_none(array, 2, 2).form
    )


def test_rpad_and_clip_regular_array():
    content = ak.contents.numpyarray.NumpyArray(
        np.array(
            [
                2.1,
                8.4,
                7.4,
                1.6,
                2.2,
                3.4,
                6.2,
                5.4,
                1.5,
                3.9,
                3.8,
                3.0,
                8.5,
                6.9,
                4.3,
                3.6,
                6.7,
                1.8,
                3.2,
            ]
        )
    )
    index = ak.index.Index64(
        np.array([13, 9, 13, 4, 8, 3, 15, -1, 16, 2, 8], dtype=np.int64)
    )
    indexedarray = ak.contents.indexedoptionarray.IndexedOptionArray(index, content)
    array = ak.contents.regulararray.RegularArray(indexedarray, 3, zeros_length=0)

    assert to_list(ak._do.pad_none(array, 5, 0, clip=True)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
        None,
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 0, clip=True).form
        == ak._do.pad_none(array, 5, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(array, 4, 0, clip=True)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 4, 0, clip=True).form
        == ak._do.pad_none(array, 4, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(array, 3, 0, clip=True)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 3, 0, clip=True).form
        == ak._do.pad_none(array, 3, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(array, 2, 0, clip=True)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 2, 0, clip=True).form
        == ak._do.pad_none(array, 2, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(array, 1, 0, clip=True)) == [[6.9, 3.9, 6.9]]
    assert (
        ak._do.pad_none(array.to_typetracer(), 1, 0, clip=True).form
        == ak._do.pad_none(array, 1, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(array, 5, 1, clip=True)) == [
        [6.9, 3.9, 6.9, None, None],
        [2.2, 1.5, 1.6, None, None],
        [3.6, None, 6.7, None, None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 1, clip=True).form
        == ak._do.pad_none(array, 5, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(array, 4, 1, clip=True)) == [
        [6.9, 3.9, 6.9, None],
        [2.2, 1.5, 1.6, None],
        [3.6, None, 6.7, None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 4, 1, clip=True).form
        == ak._do.pad_none(array, 4, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(array, 3, 1, clip=True)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 3, 1, clip=True).form
        == ak._do.pad_none(array, 3, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(array, 2, 1, clip=True)) == [
        [6.9, 3.9],
        [2.2, 1.5],
        [3.6, None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 2, 1, clip=True).form
        == ak._do.pad_none(array, 2, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(array, 1, 1, clip=True)) == [[6.9], [2.2], [3.6]]
    assert (
        ak._do.pad_none(array.to_typetracer(), 1, 1, clip=True).form
        == ak._do.pad_none(array, 1, 1, clip=True).form
    )

    array = ak.contents.numpyarray.NumpyArray(np.arange(2 * 3 * 5).reshape(2, 3, 5))
    assert to_list(array) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]

    assert to_list(ak._do.pad_none(array, 7, 2, clip=True)) == [
        [
            [0, 1, 2, 3, 4, None, None],
            [5, 6, 7, 8, 9, None, None],
            [10, 11, 12, 13, 14, None, None],
        ],
        [
            [15, 16, 17, 18, 19, None, None],
            [20, 21, 22, 23, 24, None, None],
            [25, 26, 27, 28, 29, None, None],
        ],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 7, 2, clip=True).form
        == ak._do.pad_none(array, 7, 2, clip=True).form
    )

    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)
    regulararray = ak.contents.regulararray.RegularArray(
        listoffsetarray, 2, zeros_length=0
    )

    assert to_list(ak._do.pad_none(regulararray, 1, 0, clip=True)) == [
        [[0.0, 1.1, 2.2], []]
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 1, 0, clip=True).form
        == ak._do.pad_none(regulararray, 1, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 2, 0, clip=True)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 2, 0, clip=True).form
        == ak._do.pad_none(regulararray, 2, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 3, 0, clip=True)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 3, 0, clip=True).form
        == ak._do.pad_none(regulararray, 3, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 4, 0, clip=True)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
        None,
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 4, 0, clip=True).form
        == ak._do.pad_none(regulararray, 4, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 5, 0, clip=True)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
        None,
        None,
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 5, 0, clip=True).form
        == ak._do.pad_none(regulararray, 5, 0, clip=True).form
    )

    assert to_list(ak._do.pad_none(regulararray, 1, 1, clip=True)) == [
        [[0.0, 1.1, 2.2]],
        [[3.3, 4.4]],
        [[6.6, 7.7, 8.8, 9.9]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 1, 1, clip=True).form
        == ak._do.pad_none(regulararray, 1, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 2, 1, clip=True)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 2, 1, clip=True).form
        == ak._do.pad_none(regulararray, 2, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 3, 1, clip=True)) == [
        [[0.0, 1.1, 2.2], [], None],
        [[3.3, 4.4], [5.5], None],
        [[6.6, 7.7, 8.8, 9.9], [], None],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 3, 1, clip=True).form
        == ak._do.pad_none(regulararray, 3, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 7, 1, clip=True)) == [
        [[0.0, 1.1, 2.2], [], None, None, None, None, None],
        [[3.3, 4.4], [5.5], None, None, None, None, None],
        [[6.6, 7.7, 8.8, 9.9], [], None, None, None, None, None],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 7, 1, clip=True).form
        == ak._do.pad_none(regulararray, 7, 1, clip=True).form
    )

    assert to_list(ak._do.pad_none(regulararray, 1, 2, clip=True)) == [
        [[0.0], [None]],
        [[3.3], [5.5]],
        [[6.6], [None]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 1, 2, clip=True).form
        == ak._do.pad_none(regulararray, 1, 2, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 2, 2, clip=True)) == [
        [[0.0, 1.1], [None, None]],
        [[3.3, 4.4], [5.5, None]],
        [[6.6, 7.7], [None, None]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 2, 2, clip=True).form
        == ak._do.pad_none(regulararray, 2, 2, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 3, 2, clip=True)) == [
        [[0.0, 1.1, 2.2], [None, None, None]],
        [[3.3, 4.4, None], [5.5, None, None]],
        [[6.6, 7.7, 8.8], [None, None, None]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 3, 2, clip=True).form
        == ak._do.pad_none(regulararray, 3, 2, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 4, 2, clip=True)) == [
        [[0.0, 1.1, 2.2, None], [None, None, None, None]],
        [[3.3, 4.4, None, None], [5.5, None, None, None]],
        [[6.6, 7.7, 8.8, 9.9], [None, None, None, None]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 4, 2, clip=True).form
        == ak._do.pad_none(regulararray, 4, 2, clip=True).form
    )
    assert to_list(ak._do.pad_none(regulararray, 5, 2, clip=True)) == [
        [[0.0, 1.1, 2.2, None, None], [None, None, None, None, None]],
        [[3.3, 4.4, None, None, None], [5.5, None, None, None, None]],
        [[6.6, 7.7, 8.8, 9.9, None], [None, None, None, None, None]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 5, 2, clip=True).form
        == ak._do.pad_none(regulararray, 5, 2, clip=True).form
    )


def test_rpad_regular_array():
    content = ak.contents.numpyarray.NumpyArray(
        np.array(
            [
                2.1,
                8.4,
                7.4,
                1.6,
                2.2,
                3.4,
                6.2,
                5.4,
                1.5,
                3.9,
                3.8,
                3.0,
                8.5,
                6.9,
                4.3,
                3.6,
                6.7,
                1.8,
                3.2,
            ]
        )
    )
    index = ak.index.Index64(
        np.array([13, 9, 13, 4, 8, 3, 15, -1, 16, 2, 8], dtype=np.int64)
    )
    indexedarray = ak.contents.indexedoptionarray.IndexedOptionArray(index, content)
    array = ak.contents.regulararray.RegularArray(indexedarray, 3, zeros_length=0)

    assert to_list(ak._do.pad_none(array, 5, 0)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
        None,
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 0).form
        == ak._do.pad_none(array, 5, 0).form
    )
    assert to_list(ak._do.pad_none(array, 4, 0)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 4, 0).form
        == ak._do.pad_none(array, 4, 0).form
    )
    assert to_list(ak._do.pad_none(array, 3, 0)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 3, 0).form
        == ak._do.pad_none(array, 3, 0).form
    )
    assert to_list(ak._do.pad_none(array, 1, 0)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 1, 0).form
        == ak._do.pad_none(array, 1, 0).form
    )
    assert to_list(ak._do.pad_none(array, 5, 1)) == [
        [6.9, 3.9, 6.9, None, None],
        [2.2, 1.5, 1.6, None, None],
        [3.6, None, 6.7, None, None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 1).form
        == ak._do.pad_none(array, 5, 1).form
    )
    assert to_list(ak._do.pad_none(array, 4, 1)) == [
        [6.9, 3.9, 6.9, None],
        [2.2, 1.5, 1.6, None],
        [3.6, None, 6.7, None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 4, 1).form
        == ak._do.pad_none(array, 4, 1).form
    )
    assert to_list(ak._do.pad_none(array, 3, 1)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 3, 1).form
        == ak._do.pad_none(array, 3, 1).form
    )
    assert to_list(ak._do.pad_none(array, 1, 1)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 1, 1).form
        == ak._do.pad_none(array, 1, 1).form
    )

    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)
    regulararray = ak.contents.regulararray.RegularArray(
        listoffsetarray, 2, zeros_length=0
    )

    assert to_list(ak._do.pad_none(regulararray, 1, 0)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 1, 0).form
        == ak._do.pad_none(regulararray, 1, 0).form
    )
    assert to_list(ak._do.pad_none(regulararray, 3, 0)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 3, 0).form
        == ak._do.pad_none(regulararray, 3, 0).form
    )
    assert to_list(ak._do.pad_none(regulararray, 4, 0)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
        None,
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 4, 0).form
        == ak._do.pad_none(regulararray, 4, 0).form
    )
    assert to_list(ak._do.pad_none(regulararray, 7, 0)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
        None,
        None,
        None,
        None,
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 7, 0).form
        == ak._do.pad_none(regulararray, 7, 0).form
    )

    assert to_list(ak._do.pad_none(regulararray, 1, 1)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 1, 1).form
        == ak._do.pad_none(regulararray, 1, 1).form
    )
    assert to_list(ak._do.pad_none(regulararray, 2, 1)) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 2, 1).form
        == ak._do.pad_none(regulararray, 2, 1).form
    )
    assert to_list(ak._do.pad_none(regulararray, 3, 1)) == [
        [[0.0, 1.1, 2.2], [], None],
        [[3.3, 4.4], [5.5], None],
        [[6.6, 7.7, 8.8, 9.9], [], None],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 3, 1).form
        == ak._do.pad_none(regulararray, 3, 1).form
    )
    assert to_list(ak._do.pad_none(regulararray, 5, 1)) == [
        [[0.0, 1.1, 2.2], [], None, None, None],
        [[3.3, 4.4], [5.5], None, None, None],
        [[6.6, 7.7, 8.8, 9.9], [], None, None, None],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 5, 1).form
        == ak._do.pad_none(regulararray, 5, 1).form
    )
    assert to_list(ak._do.pad_none(regulararray, 7, 1)) == [
        [[0.0, 1.1, 2.2], [], None, None, None, None, None],
        [[3.3, 4.4], [5.5], None, None, None, None, None],
        [[6.6, 7.7, 8.8, 9.9], [], None, None, None, None, None],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 7, 1).form
        == ak._do.pad_none(regulararray, 7, 1).form
    )

    assert to_list(ak._do.pad_none(regulararray, 1, 2)) == [
        [[0.0, 1.1, 2.2], [None]],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], [None]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 1, 2).form
        == ak._do.pad_none(regulararray, 1, 2).form
    )
    assert to_list(ak._do.pad_none(regulararray, 2, 2)) == [
        [[0.0, 1.1, 2.2], [None, None]],
        [[3.3, 4.4], [5.5, None]],
        [[6.6, 7.7, 8.8, 9.9], [None, None]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 2, 2).form
        == ak._do.pad_none(regulararray, 2, 2).form
    )
    assert to_list(ak._do.pad_none(regulararray, 3, 2)) == [
        [[0.0, 1.1, 2.2], [None, None, None]],
        [[3.3, 4.4, None], [5.5, None, None]],
        [[6.6, 7.7, 8.8, 9.9], [None, None, None]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 3, 2).form
        == ak._do.pad_none(regulararray, 3, 2).form
    )
    assert to_list(ak._do.pad_none(regulararray, 4, 2)) == [
        [[0.0, 1.1, 2.2, None], [None, None, None, None]],
        [[3.3, 4.4, None, None], [5.5, None, None, None]],
        [[6.6, 7.7, 8.8, 9.9], [None, None, None, None]],
    ]
    assert (
        ak._do.pad_none(regulararray.to_typetracer(), 4, 2).form
        == ak._do.pad_none(regulararray, 4, 2).form
    )


def test_rpad_and_clip_listoffset_array():
    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)
    assert to_list(listoffsetarray) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
        [],
    ]

    assert to_list(ak._do.pad_none(listoffsetarray, 3, 0, clip=True)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 3, 0, clip=True).form
        == ak._do.pad_none(listoffsetarray, 3, 0, clip=True).form
    )
    assert "option[" + str(listoffsetarray.form.type) + "]" == str(
        ak._do.pad_none(listoffsetarray, 3, 0, clip=True).form.type
    )

    assert to_list(ak._do.pad_none(listoffsetarray, 7, 0, clip=True)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
        [],
        None,
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 7, 0, clip=True).form
        == ak._do.pad_none(listoffsetarray, 7, 0, clip=True).form
    )
    assert "option[" + str(listoffsetarray.form.type) + "]" == str(
        ak._do.pad_none(listoffsetarray, 7, 0, clip=True).form.type
    )

    assert to_list(ak._do.pad_none(listoffsetarray, 5, 1, clip=True)) == [
        [0.0, 1.1, 2.2, None, None],
        [None, None, None, None, None],
        [3.3, 4.4, None, None, None],
        [5.5, None, None, None, None],
        [6.6, 7.7, 8.8, 9.9, None],
        [None, None, None, None, None],
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 5, 1, clip=True).form
        == ak._do.pad_none(listoffsetarray, 5, 1, clip=True).form
    )

    assert str(ak._do.pad_none(listoffsetarray, 5, 1).form.type) == "var * ?float64"
    assert (
        str(ak._do.pad_none(listoffsetarray, 5, 1, clip=True).form.type)
        == "5 * ?float64"
    )

    assert to_list(ak._do.pad_none(listoffsetarray, 1, 1, clip=True)) == [
        [0.0],
        [None],
        [3.3],
        [5.5],
        [6.6],
        [None],
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 1, 1, clip=True).form
        == ak._do.pad_none(listoffsetarray, 1, 1, clip=True).form
    )

    content = ak.contents.numpyarray.NumpyArray(np.array([1.5, 3.3]))
    index = ak.index.Index64(
        np.array(
            [
                0,
                -3,
                1,
                -2,
                1,
                0,
                0,
                -3,
                -13,
                0,
                1,
                1,
                0,
                1,
                1,
                1,
                1,
                -10,
                0,
                -1,
                0,
                0,
                0,
                1,
                -1,
                1,
                1,
            ]
        )
    )
    indexedarray = ak.contents.indexedoptionarray.IndexedOptionArray(index, content)
    offsets = ak.index.Index64(np.array([14, 15, 15, 15, 26, 26, 26]))
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(offsets, indexedarray)

    assert to_list(listoffsetarray) == [
        [3.3],
        [],
        [],
        [3.3, 3.3, None, 1.5, None, 1.5, 1.5, 1.5, 3.3, None, 3.3],
        [],
        [],
    ]
    assert to_list(ak._do.pad_none(listoffsetarray, 1, 0, clip=True)) == [[3.3]]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 1, 0, clip=True).form
        == ak._do.pad_none(listoffsetarray, 1, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(listoffsetarray, 1, 1, clip=True)) == [
        [3.3],
        [None],
        [None],
        [3.3],
        [None],
        [None],
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 1, 1, clip=True).form
        == ak._do.pad_none(listoffsetarray, 1, 1, clip=True).form
    )


def test_rpad_listoffset_array():
    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)

    assert to_list(listoffsetarray) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
        [],
    ]

    assert to_list(ak._do.pad_none(listoffsetarray, 3, 0)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
        [],
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 3, 0).form
        == ak._do.pad_none(listoffsetarray, 3, 0).form
    )

    assert ak.types.ArrayType(
        ak.types.OptionType(ak.types.ListType(ak.types.NumpyType("float64"))), 6
    ) == ak.operations.type(ak._do.pad_none(listoffsetarray, 3, 0))

    assert to_list(ak._do.pad_none(listoffsetarray, 7, 0)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
        [],
        None,
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 7, 0).form
        == ak._do.pad_none(listoffsetarray, 7, 0).form
    )

    assert ak.types.ArrayType(
        ak.types.OptionType(ak.types.ListType(ak.types.NumpyType("float64"))), 7
    ) == ak.operations.type(ak._do.pad_none(listoffsetarray, 7, 0))

    assert to_list(ak._do.pad_none(listoffsetarray, 5, 1)) == [
        [0.0, 1.1, 2.2, None, None],
        [None, None, None, None, None],
        [3.3, 4.4, None, None, None],
        [5.5, None, None, None, None],
        [6.6, 7.7, 8.8, 9.9, None],
        [None, None, None, None, None],
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 5, 1).form
        == ak._do.pad_none(listoffsetarray, 5, 1).form
    )
    assert ak.operations.type(
        ak._do.pad_none(listoffsetarray, 5, 1)
    ) == ak.types.ArrayType(
        ak.types.ListType(ak.types.OptionType(ak.types.NumpyType("float64"))), 6
    )

    assert to_list(ak._do.pad_none(listoffsetarray, 1, 1)) == [
        [0.0, 1.1, 2.2],
        [None],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
        [None],
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 1, 1).form
        == ak._do.pad_none(listoffsetarray, 1, 1).form
    )

    content = ak.contents.numpyarray.NumpyArray(np.array([1.5, 3.3]))
    index = ak.index.Index64(
        np.array(
            [
                0,
                -3,
                1,
                -2,
                1,
                0,
                0,
                -3,
                -13,
                0,
                1,
                1,
                0,
                1,
                1,
                1,
                1,
                -10,
                0,
                -1,
                0,
                0,
                0,
                1,
                -1,
                1,
                1,
            ]
        )
    )
    indexedarray = ak.contents.indexedoptionarray.IndexedOptionArray(index, content)
    offsets = ak.index.Index64(np.array([14, 15, 15, 15, 26, 26, 26]))
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(offsets, indexedarray)

    assert to_list(listoffsetarray) == [
        [3.3],
        [],
        [],
        [3.3, 3.3, None, 1.5, None, 1.5, 1.5, 1.5, 3.3, None, 3.3],
        [],
        [],
    ]

    assert to_list(ak._do.pad_none(listoffsetarray, 1, 0)) == [
        [3.3],
        [],
        [],
        [3.3, 3.3, None, 1.5, None, 1.5, 1.5, 1.5, 3.3, None, 3.3],
        [],
        [],
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 1, 0).form
        == ak._do.pad_none(listoffsetarray, 1, 0).form
    )
    assert ak.types.ArrayType(
        ak.types.OptionType(
            ak.types.ListType(ak.types.OptionType(ak.types.NumpyType("float64")))
        ),
        6,
    ) == ak.operations.type(ak._do.pad_none(listoffsetarray, 1, 0))

    assert to_list(ak._do.pad_none(listoffsetarray, 6, 0)) == [
        [3.3],
        [],
        [],
        [3.3, 3.3, None, 1.5, None, 1.5, 1.5, 1.5, 3.3, None, 3.3],
        [],
        [],
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 6, 0).form
        == ak._do.pad_none(listoffsetarray, 6, 0).form
    )
    assert ak.types.ArrayType(
        ak.types.OptionType(
            ak.types.ListType(ak.types.OptionType(ak.types.NumpyType("float64")))
        ),
        6,
    ) == ak.operations.type(ak._do.pad_none(listoffsetarray, 6, 0))

    assert to_list(ak._do.pad_none(listoffsetarray, 7, 0)) == [
        [3.3],
        [],
        [],
        [3.3, 3.3, None, 1.5, None, 1.5, 1.5, 1.5, 3.3, None, 3.3],
        [],
        [],
        None,
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 7, 0).form
        == ak._do.pad_none(listoffsetarray, 7, 0).form
    )
    assert ak.types.ArrayType(
        ak.types.OptionType(ak.operations.type(listoffsetarray).content), 7
    ) == ak.operations.type(ak._do.pad_none(listoffsetarray, 7, 0))

    assert to_list(ak._do.pad_none(listoffsetarray, 9, 0)) == [
        [3.3],
        [],
        [],
        [3.3, 3.3, None, 1.5, None, 1.5, 1.5, 1.5, 3.3, None, 3.3],
        [],
        [],
        None,
        None,
        None,
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 9, 0).form
        == ak._do.pad_none(listoffsetarray, 9, 0).form
    )
    assert ak.types.ArrayType(
        ak.types.OptionType(ak.operations.type(listoffsetarray).content), 9
    ) == ak.operations.type(ak._do.pad_none(listoffsetarray, 9, 0))

    assert to_list(ak._do.pad_none(listoffsetarray, 1, 1)) == [
        [3.3],
        [None],
        [None],
        [3.3, 3.3, None, 1.5, None, 1.5, 1.5, 1.5, 3.3, None, 3.3],
        [None],
        [None],
    ]
    assert (
        ak._do.pad_none(listoffsetarray.to_typetracer(), 1, 1).form
        == ak._do.pad_none(listoffsetarray, 1, 1).form
    )
    assert ak.types.ArrayType(
        ak.types.ListType(ak.types.OptionType(ak.types.NumpyType("float64"))), 6
    ) == ak.operations.type(ak._do.pad_none(listoffsetarray, 1, 1))

    assert to_list(ak._do.pad_none(listoffsetarray, 4, 1)) == [
        [3.3, None, None, None],
        [None, None, None, None],
        [None, None, None, None],
        [3.3, 3.3, None, 1.5, None, 1.5, 1.5, 1.5, 3.3, None, 3.3],
        [None, None, None, None],
        [None, None, None, None],
    ]
    assert ak.operations.type(listoffsetarray) == ak.operations.type(
        ak._do.pad_none(listoffsetarray, 4, 1)
    )


def test_rpad_list_array():
    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts = ak.index.Index64(np.array([0, 3, 4, 5, 8]))
    stops = ak.index.Index64(np.array([3, 3, 6, 8, 9]))
    array = ak.contents.ListArray(starts, stops, content)

    assert to_list(array) == [
        [0.0, 1.1, 2.2],
        [],
        [4.4, 5.5],
        [5.5, 6.6, 7.7],
        [8.8],
    ]
    assert to_list(ak._do.pad_none(array, 1, 0)) == [
        [0.0, 1.1, 2.2],
        [],
        [4.4, 5.5],
        [5.5, 6.6, 7.7],
        [8.8],
    ]
    assert ak.types.ArrayType(
        ak.types.OptionType(ak.types.ListType(ak.types.NumpyType("float64"))), 5
    ) == ak.operations.type(ak._do.pad_none(array, 1, 0))

    assert to_list(ak._do.pad_none(array, 2, 0)) == [
        [0.0, 1.1, 2.2],
        [],
        [4.4, 5.5],
        [5.5, 6.6, 7.7],
        [8.8],
    ]
    assert ak.types.ArrayType(
        ak.types.OptionType(ak.types.ListType(ak.types.NumpyType("float64"))), 5
    ) == ak.operations.type(ak._do.pad_none(array, 2, 0))

    assert to_list(ak._do.pad_none(array, 7, 0)) == [
        [0.0, 1.1, 2.2],
        [],
        [4.4, 5.5],
        [5.5, 6.6, 7.7],
        [8.8],
        None,
        None,
    ]
    assert ak.types.ArrayType(
        ak.types.OptionType(ak.types.ListType(ak.types.NumpyType("float64"))), 7
    ) == ak.operations.type(ak._do.pad_none(array, 7, 0))

    assert to_list(ak._do.pad_none(array, 1, 1)) == [
        [0.0, 1.1, 2.2],
        [None],
        [4.4, 5.5],
        [5.5, 6.6, 7.7],
        [8.8],
    ]

    assert to_list(ak._do.pad_none(array, 2, 1)) == [
        [0.0, 1.1, 2.2],
        [None, None],
        [4.4, 5.5],
        [5.5, 6.6, 7.7],
        [8.8, None],
    ]

    assert to_list(ak._do.pad_none(array, 3, 1)) == [
        [0.0, 1.1, 2.2],
        [None, None, None],
        [4.4, 5.5, None],
        [5.5, 6.6, 7.7],
        [8.8, None, None],
    ]

    assert to_list(ak._do.pad_none(array, 4, 1)) == [
        [0.0, 1.1, 2.2, None],
        [None, None, None, None],
        [4.4, 5.5, None, None],
        [5.5, 6.6, 7.7, None],
        [8.8, None, None, None],
    ]


def test_rpad_and_clip_list_array():
    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts = ak.index.Index64(np.array([0, 3, 4, 5, 8]))
    stops = ak.index.Index64(np.array([3, 3, 6, 8, 9]))
    array = ak.contents.listarray.ListArray(starts, stops, content)

    assert to_list(array) == [
        [0.0, 1.1, 2.2],
        [],
        [4.4, 5.5],
        [5.5, 6.6, 7.7],
        [8.8],
    ]
    assert to_list(ak._do.pad_none(array, 1, 0, clip=True)) == [[0.0, 1.1, 2.2]]
    assert (
        ak._do.pad_none(array.to_typetracer(), 1, 0, clip=True).form
        == ak._do.pad_none(array, 1, 0, clip=True).form
    )
    assert ak.types.ArrayType(
        ak.types.OptionType(ak.types.ListType(ak.types.NumpyType("float64"))), 1
    ) == ak.operations.type(ak._do.pad_none(array, 1, 0, clip=True))

    assert to_list(ak._do.pad_none(array, 2, 0, clip=True)) == [[0.0, 1.1, 2.2], []]
    assert (
        ak._do.pad_none(array.to_typetracer(), 2, 0, clip=True).form
        == ak._do.pad_none(array, 2, 0, clip=True).form
    )
    assert ak.types.ArrayType(
        ak.types.OptionType(ak.types.ListType(ak.types.NumpyType("float64"))), 2
    ) == ak.operations.type(ak._do.pad_none(array, 2, 0, clip=True))

    assert to_list(ak._do.pad_none(array, 7, 0, clip=True)) == [
        [0.0, 1.1, 2.2],
        [],
        [4.4, 5.5],
        [5.5, 6.6, 7.7],
        [8.8],
        None,
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 7, 0, clip=True).form
        == ak._do.pad_none(array, 7, 0, clip=True).form
    )
    assert ak.types.ArrayType(
        ak.types.OptionType(ak.types.ListType(ak.types.NumpyType("float64"))), 7
    ) == ak.operations.type(ak._do.pad_none(array, 7, 0, clip=True))

    assert to_list(ak._do.pad_none(array, 1, 1, clip=True)) == [
        [0.0],
        [None],
        [4.4],
        [5.5],
        [8.8],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 1, 1, clip=True).form
        == ak._do.pad_none(array, 1, 1, clip=True).form
    )

    assert to_list(ak._do.pad_none(array, 2, 1, clip=True)) == [
        [0.0, 1.1],
        [None, None],
        [4.4, 5.5],
        [5.5, 6.6],
        [8.8, None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 2, 1, clip=True).form
        == ak._do.pad_none(array, 2, 1, clip=True).form
    )


def test_rpad_indexed_array():
    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts = ak.index.Index64(np.array([0, 3, 3, 5, 6]))
    stops = ak.index.Index64(np.array([3, 3, 5, 6, 10]))
    listoffsetarray = ak.contents.listarray.ListArray(starts, stops, content)
    content = ak.contents.numpyarray.NumpyArray(
        np.array([6.6, 7.7, 8.8, 9.9, 5.5, 3.3, 4.4, 0.0, 1.1, 2.2])
    )
    offsets = ak.index.Index64(np.array([0, 4, 5, 7, 7, 10]))
    backward = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)

    index = ak.index.Index64(np.array([4, 3, 2, 1, 0], dtype=np.int64))
    indexedarray = ak.contents.indexedarray.IndexedArray(index, listoffsetarray)
    assert to_list(indexedarray) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
    ]

    assert to_list(ak._do.pad_none(backward, 4, 1)) == to_list(
        ak._do.pad_none(indexedarray, 4, 1)
    )
    assert to_list(ak._do.pad_none(indexedarray, 1, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
    ]
    assert to_list(ak._do.pad_none(indexedarray, 2, 1)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5, None],
        [3.3, 4.4],
        [None, None],
        [0.0, 1.1, 2.2],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 2, 1).form
        == ak._do.pad_none(indexedarray, 2, 1).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 3, 1)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5, None, None],
        [3.3, 4.4, None],
        [None, None, None],
        [0.0, 1.1, 2.2],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 3, 1).form
        == ak._do.pad_none(indexedarray, 3, 1).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 4, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 4, 0).form
        == ak._do.pad_none(indexedarray, 4, 0).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 5, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 5, 0).form
        == ak._do.pad_none(indexedarray, 5, 0).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 6, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
        None,
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 6, 0).form
        == ak._do.pad_none(indexedarray, 6, 0).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 7, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
        None,
        None,
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 7, 0).form
        == ak._do.pad_none(indexedarray, 7, 0).form
    )


def test_rpad_and_clip_indexed_array():
    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 10]))
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)
    content = ak.contents.numpyarray.NumpyArray(
        np.array([6.6, 7.7, 8.8, 9.9, 5.5, 3.3, 4.4, 0.0, 1.1, 2.2])
    )
    offsets = ak.index.Index64(np.array([0, 4, 5, 7, 7, 10]))
    backward = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)

    index = ak.index.Index64(np.array([4, 3, 2, 1, 0], dtype=np.int64))
    indexedarray = ak.contents.IndexedArray(index, listoffsetarray)
    assert to_list(indexedarray) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
    ]

    assert to_list(ak._do.pad_none(backward, 4, 1, clip=True)) == to_list(
        ak._do.pad_none(indexedarray, 4, 1, clip=True)
    )
    assert to_list(ak._do.pad_none(indexedarray, 1, 0, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9]
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 1, 0, clip=True).form
        == ak._do.pad_none(indexedarray, 1, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 2, 0, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 2, 0, clip=True).form
        == ak._do.pad_none(indexedarray, 2, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 3, 0, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 3, 0, clip=True).form
        == ak._do.pad_none(indexedarray, 3, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 4, 0, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 4, 0, clip=True).form
        == ak._do.pad_none(indexedarray, 4, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 5, 0, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 5, 0, clip=True).form
        == ak._do.pad_none(indexedarray, 5, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 6, 0, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
        None,
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 6, 0, clip=True).form
        == ak._do.pad_none(indexedarray, 6, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 7, 0, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
        None,
        None,
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 7, 0, clip=True).form
        == ak._do.pad_none(indexedarray, 7, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 8, 0, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [],
        [0.0, 1.1, 2.2],
        None,
        None,
        None,
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 8, 0, clip=True).form
        == ak._do.pad_none(indexedarray, 8, 0, clip=True).form
    )

    assert to_list(ak._do.pad_none(indexedarray, 1, 1, clip=True)) == [
        [6.6],
        [5.5],
        [3.3],
        [None],
        [0.0],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 1, 1, clip=True).form
        == ak._do.pad_none(indexedarray, 1, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 2, 1, clip=True)) == [
        [6.6, 7.7],
        [5.5, None],
        [3.3, 4.4],
        [None, None],
        [0.0, 1.1],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 2, 1, clip=True).form
        == ak._do.pad_none(indexedarray, 2, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 3, 1, clip=True)) == [
        [6.6, 7.7, 8.8],
        [5.5, None, None],
        [3.3, 4.4, None],
        [None, None, None],
        [0.0, 1.1, 2.2],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 3, 1, clip=True).form
        == ak._do.pad_none(indexedarray, 3, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 4, 1, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5, None, None, None],
        [3.3, 4.4, None, None],
        [None, None, None, None],
        [0.0, 1.1, 2.2, None],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 4, 1, clip=True).form
        == ak._do.pad_none(indexedarray, 4, 1, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 5, 1, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9, None],
        [5.5, None, None, None, None],
        [3.3, 4.4, None, None, None],
        [None, None, None, None, None],
        [0.0, 1.1, 2.2, None, None],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 5, 1, clip=True).form
        == ak._do.pad_none(indexedarray, 5, 1, clip=True).form
    )


def test_rpad_indexed_option_array():
    content = ak.contents.numpyarray.NumpyArray(
        np.asarray([0.0, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    index = ak.index.Index64(np.asarray([0, -1, -1, 1, 2, 3, 4, 5, 6, 7]))
    offsets = ak.index.Index64(np.asarray([0, 3, 5, 6, 10]))
    indexedarray = ak.contents.indexedoptionarray.IndexedOptionArray(index, content)
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(offsets, indexedarray)
    index = ak.index.Index64(
        np.asarray(
            [
                0,
                -1,
                1,
                2,
                3,
            ]
        )
    )
    indexedoptionarray = ak.contents.indexedoptionarray.IndexedOptionArray(
        index, listoffsetarray
    )
    content = ak.contents.numpyarray.NumpyArray(
        np.asarray([6.6, 7.7, 8.8, 9.9, 5.5, 3.3, 4.4, 0.0])
    )
    index = ak.index.Index64(np.asarray([0, 1, 2, 3, 4, 5, 6, 7, -1, -1]))
    offsets = ak.index.Index64(np.asarray([0, 4, 5, 7, 10]))
    indexedarray = ak.contents.indexedoptionarray.IndexedOptionArray(index, content)
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(offsets, indexedarray)
    index = ak.index.Index64(
        np.asarray(
            [
                0,
                1,
                2,
                -1,
                3,
            ]
        )
    )
    backward = ak.contents.indexedoptionarray.IndexedOptionArray(index, listoffsetarray)

    index = ak.index.Index64(np.array([4, 3, 2, -1, 0], dtype=np.int64))
    indexedarray = ak.contents.indexedoptionarray.IndexedOptionArray.simplified(
        index, indexedoptionarray
    )
    assert to_list(indexedarray) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        None,
        [0.0, None, None],
    ]

    assert to_list(ak._do.pad_none(backward, 4, 1)) == to_list(
        ak._do.pad_none(indexedarray, 4, 1)
    )
    assert to_list(ak._do.pad_none(indexedarray, 1, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        None,
        [0.0, None, None],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 1, 0).form
        == ak._do.pad_none(indexedarray, 1, 0).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 1, 1)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        None,
        [0.0, None, None],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 1, 1).form
        == ak._do.pad_none(indexedarray, 1, 1).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 3, 1)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5, None, None],
        [3.3, 4.4, None],
        None,
        [0.0, None, None],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 3, 1).form
        == ak._do.pad_none(indexedarray, 3, 1).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 4, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        None,
        [0.0, None, None],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 4, 0).form
        == ak._do.pad_none(indexedarray, 4, 0).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 5, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        None,
        [0.0, None, None],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 5, 0).form
        == ak._do.pad_none(indexedarray, 5, 0).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 6, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        None,
        [0.0, None, None],
        None,
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 6, 0).form
        == ak._do.pad_none(indexedarray, 6, 0).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 7, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        None,
        [0.0, None, None],
        None,
        None,
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 7, 0).form
        == ak._do.pad_none(indexedarray, 7, 0).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 8, 0)) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        None,
        [0.0, None, None],
        None,
        None,
        None,
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 8, 0).form
        == ak._do.pad_none(indexedarray, 8, 0).form
    )

    assert to_list(ak._do.pad_none(indexedarray, 1, 0, clip=True)) == [
        [6.6, 7.7, 8.8, 9.9]
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 1, 0, clip=True).form
        == ak._do.pad_none(indexedarray, 1, 0, clip=True).form
    )
    assert to_list(ak._do.pad_none(indexedarray, 1, 1, clip=True)) == [
        [6.6],
        [5.5],
        [3.3],
        None,
        [0.0],
    ]
    assert (
        ak._do.pad_none(indexedarray.to_typetracer(), 1, 1, clip=True).form
        == ak._do.pad_none(indexedarray, 1, 1, clip=True).form
    )


def test_rpad_recordarray():
    keys = ["x", "y"]
    offsets = ak.index.Index64(np.asarray([0, 0, 1, 3]))
    content = ak.contents.numpyarray.NumpyArray(np.asarray([1.1, 2.2, 2.2]))
    content1 = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)
    offsets = ak.index.Index64(np.asarray([0, 2, 3, 3]))
    content = ak.contents.numpyarray.NumpyArray(np.asarray([2, 2, 1]))
    content2 = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)
    contents = [content1, content2]
    array = ak.contents.recordarray.RecordArray(contents, keys)

    assert to_list(ak._do.pad_none(array, 5, 0)) == [
        {"x": [], "y": [2, 2]},
        {"x": [1.1], "y": [1]},
        {"x": [2.2, 2.2], "y": []},
        None,
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 5, 0).form
        == ak._do.pad_none(array, 5, 0).form
    )

    assert to_list(ak._do.pad_none(array, 2, 1)) == [
        {"x": [None, None], "y": [2, 2]},
        {"x": [1.1, None], "y": [1, None]},
        {"x": [2.2, 2.2], "y": [None, None]},
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 2, 1).form
        == ak._do.pad_none(array, 2, 1).form
    )


def test_rpad_unionarray():
    offsets = ak.index.Index64(np.asarray([0, 0, 1, 3]))
    content = ak.contents.numpyarray.NumpyArray(np.asarray([1.1, 2.2, 2.2]))
    content1 = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)
    offsets = ak.index.Index64(np.asarray([0, 2, 3, 3]))
    content = ak.from_iter(["2", "2", "1"], highlevel=False)
    content2 = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)
    tags = ak.index.Index8(np.array([0, 1, 0, 1, 0, 1], dtype=np.int8))
    index = ak.index.Index64(np.array([0, 0, 1, 1, 2, 2], dtype=np.int64))
    array = ak.contents.unionarray.UnionArray(tags, index, [content1, content2])
    assert to_list(array) == [[], ["2", "2"], [1.1], ["1"], [2.2, 2.2], []]

    assert to_list(ak._do.pad_none(array, 7, 0)) == [
        [],
        ["2", "2"],
        [1.1],
        ["1"],
        [2.2, 2.2],
        [],
        None,
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 7, 0).form
        == ak._do.pad_none(array, 7, 0).form
    )

    assert to_list(ak._do.pad_none(array, 2, 1)) == [
        [None, None],
        ["2", "2"],
        [1.1, None],
        ["1", None],
        [2.2, 2.2],
        [None, None],
    ]
    assert (
        ak._do.pad_none(array.to_typetracer(), 2, 1).form
        == ak._do.pad_none(array, 2, 1).form
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest

import awkward as ak

jax = pytest.importorskip("jax")
jax.config.update("jax_platform_name", "cpu")
jax.config.update("jax_enable_x64", True)

ak.jax.register_and_check()


@pytest.mark.skip("Jacobian support not implemented")
def test():
    array = ak.Array([[1, 2, 3], [4, 5, 6.0]])

    def func(x):
        return x * 2 - 1

    array_np = ak.to_numpy(array)
    jac_np = jax.jacfwd(func)(array_np)

    jac = jax.jacfwd(func)(array)
    assert jac.to_list() == jac_np.to_list()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    x = ak.from_numpy(np.empty((5, 0), dtype=np.float64), regulararray=True)
    y = ak.from_numpy(np.empty((8, 0), dtype=np.float64), regulararray=True)
    z = ak.concatenate([x, y], axis=0)
    assert len(z) == 13



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

jax = pytest.importorskip("jax")
jax.config.update("jax_platform_name", "cpu")
jax.config.update("jax_enable_x64", True)

ak.jax.register_and_check()


class GradBehavior(ak.Array):
    ...


def test():
    def square_fifth_entry(x):
        return x[4] ** 2

    primal = ak.Array(np.arange(10, dtype=np.float64), backend="jax")
    tangent = ak.Array(np.arange(10, dtype=np.float64), backend="jax")

    behavior = {"grad": GradBehavior}
    primal_grad = ak.with_parameter(primal, "__array__", "grad", behavior=behavior)
    tangent_grad = ak.with_parameter(tangent, "__array__", "grad", behavior=behavior)
    value_jvp_grad, jvp_grad_grad = jax.jvp(
        square_fifth_entry, (primal_grad,), (tangent_grad,)
    )

    assert value_jvp_grad == pytest.approx(16.0)
    assert jvp_grad_grad == pytest.approx(32.0)


def test_jvp_nested_list():
    # with jax.checking_leaks():
    array = ak.Array(np.array([[1.0, 2.0, 3.0, 4.0, 5.0]]), backend="jax")
    tangent = ak.Array(np.array([[0.0, 0.0, 0.0, 0.0, 1.0]]), backend="jax")

    def func(x):
        return x[::-1] ** 2

    # TODO: enable leak checing
    # with jax.checking_leaks():
    value_jvp, jvp_grad = jax.jvp(func, (array,), (tangent,))
    assert value_jvp.to_list() == [[1.0, 4.0, 9.0, 16.0, 25.0]]
    assert jvp_grad.to_list() == [[0.0, 0.0, 0.0, 0.0, 10.0]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numba
import numba.core.typing
import numba.core.typing.ctypes_utils
import numpy
from awkward_cpp import libawkward

import awkward as ak

dynamic_addrs = {}


def globalstring(context, builder, pyvalue):
    import llvmlite.ir.types

    if pyvalue not in dynamic_addrs:
        buf = dynamic_addrs[pyvalue] = numpy.array(pyvalue.encode("utf-8") + b"\x00")
        context.add_dynamic_addr(builder, buf.ctypes.data, info=f"str({pyvalue!r})")
    ptr = context.get_constant(numba.types.uintp, dynamic_addrs[pyvalue].ctypes.data)
    return builder.inttoptr(ptr, llvmlite.ir.PointerType(llvmlite.ir.IntType(8)))


class ArrayBuilderType(numba.types.Type):
    def __init__(self, behavior):
        super().__init__(
            name="ak.ArrayBuilderType({})".format(
                ak._connect.numba.arrayview.repr_behavior(behavior)
            )
        )
        self.behavior = behavior


@numba.extending.register_model(ArrayBuilderType)
class ArrayBuilderModel(numba.core.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("rawptr", numba.types.voidptr), ("pyptr", numba.types.pyobject)]
        super().__init__(dmm, fe_type, members)


@numba.core.imputils.lower_constant(ArrayBuilderType)
def lower_const_ArrayBuilder(context, builder, arraybuildertype, arraybuilder):
    layout = arraybuilder._layout
    rawptr = context.get_constant(numba.intp, arraybuilder._layout._ptr)
    proxyout = context.make_helper(builder, arraybuildertype)
    proxyout.rawptr = builder.inttoptr(
        rawptr, context.get_value_type(numba.types.voidptr)
    )
    proxyout.pyptr = context.add_dynamic_addr(
        builder, id(layout), info=str(type(layout))
    )
    return proxyout._getvalue()


@numba.extending.unbox(ArrayBuilderType)
def unbox_ArrayBuilder(arraybuildertype, arraybuilderobj, c):
    inner_obj = c.pyapi.object_getattr_string(arraybuilderobj, "_layout")
    rawptr_obj = c.pyapi.object_getattr_string(inner_obj, "_ptr")

    proxyout = c.context.make_helper(c.builder, arraybuildertype)
    proxyout.rawptr = c.pyapi.long_as_voidptr(rawptr_obj)
    proxyout.pyptr = inner_obj

    c.pyapi.decref(inner_obj)
    c.pyapi.decref(rawptr_obj)

    is_error = numba.core.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)


@numba.extending.box(ArrayBuilderType)
def box_ArrayBuilder(arraybuildertype, arraybuilderval, c):
    ArrayBuilder_obj = c.pyapi.unserialize(
        c.pyapi.serialize_object(ak.highlevel.ArrayBuilder)
    )
    behavior_obj = c.pyapi.unserialize(
        c.pyapi.serialize_object(arraybuildertype.behavior)
    )

    proxyin = c.context.make_helper(c.builder, arraybuildertype, arraybuilderval)
    c.pyapi.incref(proxyin.pyptr)

    out = c.pyapi.call_method(ArrayBuilder_obj, "_wrap", (proxyin.pyptr, behavior_obj))

    c.pyapi.decref(ArrayBuilder_obj)
    c.pyapi.decref(behavior_obj)
    c.pyapi.decref(proxyin.pyptr)

    return out


def call(context, builder, fcn, args):
    numbatype = numba.core.typing.ctypes_utils.make_function_type(fcn)
    fcntype = context.get_function_pointer_type(numbatype)
    fcnval = context.add_dynamic_addr(
        builder, numbatype.get_pointer(fcn), info=fcn.name
    )
    fcnptr = builder.bitcast(fcnval, fcntype)
    err = context.call_function_pointer(builder, fcnptr, args)
    with builder.if_then(
        builder.icmp_unsigned("!=", err, context.get_constant(numba.uint8, 0)),
        likely=False,
    ):
        context.call_conv.return_user_exc(builder, ValueError, (fcn.name + " failed",))


@numba.core.typing.templates.infer_global(len)
class type_len(numba.core.typing.templates.AbstractTemplate):
    def generic(self, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], ArrayBuilderType)
        ):
            return numba.intp(args[0])


@numba.extending.lower_builtin(len, ArrayBuilderType)
def lower_len(context, builder, sig, args):
    (arraybuildertype,) = sig.args
    (arraybuilderval,) = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    result = numba.core.cgutils.alloca_once(
        builder, context.get_value_type(numba.int64)
    )
    call(
        context,
        builder,
        libawkward.ArrayBuilder_length,
        (proxyin.rawptr, result),
    )
    return ak._connect.numba.layout.castint(
        context, builder, numba.int64, numba.intp, builder.load(result)
    )


@numba.core.typing.templates.infer_getattr
class type_methods(numba.core.typing.templates.AttributeTemplate):
    key = ArrayBuilderType

    @numba.core.typing.templates.bound_function("clear")
    def resolve_clear(self, arraybuildertype, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for ArrayBuilder.clear")

    @numba.core.typing.templates.bound_function("null")
    def resolve_null(self, arraybuildertype, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for ArrayBuilder.null")

    @numba.core.typing.templates.bound_function("boolean")
    def resolve_boolean(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], numba.types.Boolean)
        ):
            return numba.types.none(args[0])
        else:
            raise TypeError(
                "wrong number or types of arguments for ArrayBuilder.boolean"
            )

    @numba.core.typing.templates.bound_function("integer")
    def resolve_integer(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], numba.types.Integer)
        ):
            return numba.types.none(args[0])
        else:
            raise TypeError(
                "wrong number or types of arguments for ArrayBuilder.integer"
            )

    @numba.core.typing.templates.bound_function("real")
    def resolve_real(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], (numba.types.Integer, numba.types.Float))
        ):
            return numba.types.none(args[0])
        else:
            raise TypeError("wrong number or types of arguments for ArrayBuilder.real")

    @numba.core.typing.templates.bound_function("complex")
    def resolve_complex(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(
                args[0], (numba.types.Integer, numba.types.Float, numba.types.Complex)
            )
        ):
            return numba.types.none(args[0])
        else:
            raise TypeError(
                "wrong number or types of arguments for ArrayBuilder.complex"
            )

    @numba.core.typing.templates.bound_function("datetime")
    def resolve_datetime(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], (numba.types.NPDatetime, numba.types.UnicodeType))
        ):
            return numba.types.none(args[0])
        else:
            raise TypeError(
                "wrong number or types of arguments for ArrayBuilder.datetime"
            )

    @numba.core.typing.templates.bound_function("timedelta")
    def resolve_timedelta(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], (numba.types.NPTimedelta, numba.types.UnicodeType))
        ):
            return numba.types.none(args[0])
        else:
            raise TypeError(
                "wrong number or types of arguments for ArrayBuilder.timedelta"
            )

    @numba.core.typing.templates.bound_function("string")
    def resolve_string(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], (numba.types.UnicodeType))
        ):
            return numba.types.none(args[0])
        else:
            raise TypeError(
                "wrong number or types of arguments for ArrayBuilder.string"
            )

    @numba.core.typing.templates.bound_function("begin_list")
    def resolve_begin_list(self, arraybuildertype, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for ArrayBuilder.begin_list")

    @numba.core.typing.templates.bound_function("end_list")
    def resolve_end_list(self, arraybuildertype, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for ArrayBuilder.end_list")

    @numba.core.typing.templates.bound_function("begin_tuple")
    def resolve_begin_tuple(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], numba.types.Integer)
        ):
            return numba.types.none(args[0])
        else:
            raise TypeError(
                "wrong number or types of arguments for ArrayBuilder.begin_tuple"
            )

    @numba.core.typing.templates.bound_function("index")
    def resolve_index(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], numba.types.Integer)
        ):
            return arraybuildertype(args[0])
        else:
            raise TypeError("wrong number or types of arguments for ArrayBuilder.index")

    @numba.core.typing.templates.bound_function("end_tuple")
    def resolve_end_tuple(self, arraybuildertype, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for ArrayBuilder.end_tuple")

    @numba.core.typing.templates.bound_function("begin_record")
    def resolve_begin_record(self, arraybuildertype, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        elif (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], numba.types.StringLiteral)
        ):
            return numba.types.none(args[0])
        else:
            raise TypeError(
                "wrong number or types of arguments for ArrayBuilder.begin_record"
            )

    @numba.core.typing.templates.bound_function("field")
    def resolve_field(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], numba.types.StringLiteral)
        ):
            return arraybuildertype(args[0])
        else:
            raise TypeError("wrong number or types of arguments for ArrayBuilder.field")

    @numba.core.typing.templates.bound_function("end_record")
    def resolve_end_record(self, arraybuildertype, args, kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return numba.types.none()
        else:
            raise TypeError("wrong number of arguments for ArrayBuilder.end_record")

    @numba.core.typing.templates.bound_function("append")
    def resolve_append(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(
                args[0],
                (
                    ak._connect.numba.arrayview.ArrayViewType,
                    ak._connect.numba.arrayview.RecordViewType,
                    numba.types.Boolean,
                    numba.types.Integer,
                    numba.types.Float,
                    numba.types.Complex,
                    numba.types.NPDatetime,
                    numba.types.NPTimedelta,
                    numba.types.UnicodeType,
                ),
            )
        ):
            return numba.types.none(args[0])
        elif (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], numba.types.Optional)
            and isinstance(
                args[0].type,
                (
                    numba.types.Boolean,
                    numba.types.Integer,
                    numba.types.Float,
                    numba.types.Complex,
                    numba.types.NPDatetime,
                    numba.types.NPTimedelta,
                    numba.types.UnicodeType,
                ),
            )
        ):
            return numba.types.none(args[0])
        elif (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], numba.types.NoneType)
        ):
            return numba.types.none(args[0])
        else:
            if len(args) == 1 and arraybuildertype.behavior is not None:
                for key, lower in arraybuildertype.behavior.items():
                    if (
                        isinstance(key, tuple)
                        and len(key) == 3
                        and key[0] == "__numba_lower__"
                        and key[1] == ak.highlevel.ArrayBuilder.append
                        and (
                            args[0] == key[2]
                            or (
                                isinstance(key[2], type) and isinstance(args[0], key[2])
                            )
                        )
                    ):
                        numba.extending.lower_builtin(
                            "append", ArrayBuilderType, args[0]
                        )(lower)
                        return numba.types.none(args[0])

            raise TypeError(
                "wrong number or types of arguments for ArrayBuilder.append"
            )

    @numba.core.typing.templates.bound_function("extend")
    def resolve_extend(self, arraybuildertype, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], ak._connect.numba.arrayview.ArrayViewType)
        ):
            return numba.types.none(args[0])
        else:
            raise TypeError(
                "wrong number or types of arguments for ArrayBuilder.extend"
            )


@numba.extending.lower_builtin("clear", ArrayBuilderType)
def lower_clear(context, builder, sig, args):
    (arraybuildertype,) = sig.args
    (arraybuilderval,) = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    call(context, builder, libawkward.ArrayBuilder_clear, (proxyin.rawptr,))
    return context.get_dummy_value()


@numba.extending.lower_builtin("null", ArrayBuilderType)
def lower_null(context, builder, sig, args):
    (arraybuildertype,) = sig.args
    (arraybuilderval,) = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    call(context, builder, libawkward.ArrayBuilder_null, (proxyin.rawptr,))
    return context.get_dummy_value()


@numba.extending.lower_builtin("boolean", ArrayBuilderType, numba.types.Boolean)
def lower_boolean(context, builder, sig, args):
    arraybuildertype, xtype = sig.args
    arraybuilderval, xval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    x = builder.zext(xval, context.get_value_type(numba.uint8))
    call(context, builder, libawkward.ArrayBuilder_boolean, (proxyin.rawptr, x))
    return context.get_dummy_value()


@numba.extending.lower_builtin("integer", ArrayBuilderType, numba.types.Integer)
def lower_integer(context, builder, sig, args):
    arraybuildertype, xtype = sig.args
    arraybuilderval, xval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    x = ak._connect.numba.layout.castint(context, builder, xtype, numba.int64, xval)
    call(context, builder, libawkward.ArrayBuilder_integer, (proxyin.rawptr, x))
    return context.get_dummy_value()


@numba.extending.lower_builtin("real", ArrayBuilderType, numba.types.Integer)
@numba.extending.lower_builtin("real", ArrayBuilderType, numba.types.Float)
def lower_real(context, builder, sig, args):
    arraybuildertype, xtype = sig.args
    arraybuilderval, xval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    if isinstance(xtype, numba.types.Integer) and xtype.signed:
        x = builder.sitofp(xval, context.get_value_type(numba.types.float64))
    elif isinstance(xtype, numba.types.Integer):
        x = builder.uitofp(xval, context.get_value_type(numba.types.float64))
    elif xtype.bitwidth < 64:
        x = builder.fpext(xval, context.get_value_type(numba.types.float64))
    elif xtype.bitwidth > 64:
        x = builder.fptrunc(xval, context.get_value_type(numba.types.float64))
    else:
        x = xval
    call(context, builder, libawkward.ArrayBuilder_real, (proxyin.rawptr, x))
    return context.get_dummy_value()


@numba.extending.lower_builtin("complex", ArrayBuilderType, numba.types.Integer)
@numba.extending.lower_builtin("complex", ArrayBuilderType, numba.types.Float)
def lower_complex_from_integer_or_float(context, builder, sig, args):
    arraybuildertype, xtype = sig.args
    arraybuilderval, xval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)

    if isinstance(xtype, numba.types.Integer) and xtype.signed:
        z_real = builder.sitofp(xval, context.get_value_type(numba.types.float64))
        z_imag = z_real.type(0)
    elif isinstance(xtype, numba.types.Integer):
        z_real = builder.uitofp(xval, context.get_value_type(numba.types.float64))
        z_imag = z_real.type(0)
    elif xtype.bitwidth < 64:
        z_real = builder.fpext(xval, context.get_value_type(numba.types.float64))
    elif xtype.bitwidth > 64:
        z_real = builder.fptrunc(xval, context.get_value_type(numba.types.float64))
    else:
        z_real = xval
    z_imag = z_real.type(0)

    call(
        context,
        builder,
        libawkward.ArrayBuilder_complex,
        (proxyin.rawptr, z_real, z_imag),
    )
    return context.get_dummy_value()


@numba.extending.lower_builtin("complex", ArrayBuilderType, numba.types.Complex)
def lower_complex(context, builder, sig, args):
    arraybuildertype, xtype = sig.args
    arraybuilderval, xval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)

    z = context.make_complex(builder, xtype, xval)
    z_real, z_imag = z.real, z.imag

    call(
        context,
        builder,
        libawkward.ArrayBuilder_complex,
        (proxyin.rawptr, z_real, z_imag),
    )
    return context.get_dummy_value()


@numba.extending.lower_builtin("datetime", ArrayBuilderType, numba.types.NPDatetime)
def lower_datetime(context, builder, sig, args):
    arraybuildertype, xtype = sig.args
    arraybuilderval, xval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    unit = globalstring(context, builder, f"datetime64[{xtype.unit}]")
    call(
        context,
        builder,
        libawkward.ArrayBuilder_datetime,
        (proxyin.rawptr, xval, unit),
    )
    return context.get_dummy_value()


@numba.extending.lower_builtin("timedelta", ArrayBuilderType, numba.types.NPTimedelta)
def lower_timedelta(context, builder, sig, args):
    arraybuildertype, xtype = sig.args
    arraybuilderval, xval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    unit = globalstring(context, builder, f"timedelta64[{xtype.unit}]")
    call(
        context,
        builder,
        libawkward.ArrayBuilder_timedelta,
        (proxyin.rawptr, xval, unit),
    )
    return context.get_dummy_value()


@numba.extending.lower_builtin("string", ArrayBuilderType, numba.types.UnicodeType)
def lower_string(context, builder, sig, args):
    arraybuildertype, xtype = sig.args
    arraybuilderval, xval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)

    pyapi = context.get_python_api(builder)
    gil = pyapi.gil_ensure()

    is_ok, out, length = pyapi.string_as_string_and_size(xval.value)
    length = ak._connect.numba.layout.castint(
        context, builder, numba.ssize_t, numba.int64, length
    )
    call(
        context,
        builder,
        libawkward.ArrayBuilder_string_length,
        (proxyin.rawptr, out, length),
    )

    pyapi.gil_release(gil)

    return context.get_dummy_value()


@numba.extending.lower_builtin("begin_list", ArrayBuilderType)
def lower_beginlist(context, builder, sig, args):
    (arraybuildertype,) = sig.args
    (arraybuilderval,) = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    call(context, builder, libawkward.ArrayBuilder_beginlist, (proxyin.rawptr,))
    return context.get_dummy_value()


@numba.extending.lower_builtin("end_list", ArrayBuilderType)
def lower_endlist(context, builder, sig, args):
    (arraybuildertype,) = sig.args
    (arraybuilderval,) = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    call(context, builder, libawkward.ArrayBuilder_endlist, (proxyin.rawptr,))
    return context.get_dummy_value()


@numba.extending.lower_builtin("begin_tuple", ArrayBuilderType, numba.types.Integer)
def lower_begintuple(context, builder, sig, args):
    arraybuildertype, numfieldstype = sig.args
    arraybuilderval, numfieldsval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    numfields = ak._connect.numba.layout.castint(
        context, builder, numfieldstype, numba.int64, numfieldsval
    )
    call(
        context,
        builder,
        libawkward.ArrayBuilder_begintuple,
        (proxyin.rawptr, numfields),
    )
    return context.get_dummy_value()


@numba.extending.lower_builtin("index", ArrayBuilderType, numba.types.Integer)
def lower_index(context, builder, sig, args):
    arraybuildertype, indextype = sig.args
    arraybuilderval, indexval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    index = ak._connect.numba.layout.castint(
        context, builder, indextype, numba.int64, indexval
    )
    call(
        context,
        builder,
        libawkward.ArrayBuilder_index,
        (proxyin.rawptr, index),
    )
    return arraybuilderval


@numba.extending.lower_builtin("end_tuple", ArrayBuilderType)
def lower_endtuple(context, builder, sig, args):
    (arraybuildertype,) = sig.args
    (arraybuilderval,) = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    call(context, builder, libawkward.ArrayBuilder_endtuple, (proxyin.rawptr,))
    return context.get_dummy_value()


@numba.extending.lower_builtin("begin_record", ArrayBuilderType)
def lower_beginrecord(context, builder, sig, args):
    (arraybuildertype,) = sig.args
    (arraybuilderval,) = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    call(
        context,
        builder,
        libawkward.ArrayBuilder_beginrecord,
        (proxyin.rawptr,),
    )
    return context.get_dummy_value()


@numba.extending.lower_builtin(
    "begin_record", ArrayBuilderType, numba.types.StringLiteral
)
def lower_beginrecord_field(context, builder, sig, args):
    arraybuildertype, nametype = sig.args
    arraybuilderval, nameval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    name = globalstring(context, builder, nametype.literal_value)
    call(
        context,
        builder,
        libawkward.ArrayBuilder_beginrecord_fast,
        (proxyin.rawptr, name),
    )
    return context.get_dummy_value()


@numba.extending.lower_builtin("field", ArrayBuilderType, numba.types.StringLiteral)
def lower_field(context, builder, sig, args):
    arraybuildertype, keytype = sig.args
    arraybuilderval, keyval = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    key = globalstring(context, builder, keytype.literal_value)
    call(
        context,
        builder,
        libawkward.ArrayBuilder_field_fast,
        (proxyin.rawptr, key),
    )
    return arraybuilderval


@numba.extending.lower_builtin("end_record", ArrayBuilderType)
def lower_endrecord(context, builder, sig, args):
    (arraybuildertype,) = sig.args
    (arraybuilderval,) = args
    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    call(context, builder, libawkward.ArrayBuilder_endrecord, (proxyin.rawptr,))
    return context.get_dummy_value()


@numba.extending.lower_builtin(
    "append", ArrayBuilderType, ak._connect.numba.arrayview.ArrayViewType
)
def lower_append_array(context, builder, sig, args):
    arraybuildertype, viewtype = sig.args
    arraybuilderval, viewval = args

    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    call(context, builder, libawkward.ArrayBuilder_beginlist, (proxyin.rawptr,))

    lower_extend_array(context, builder, sig, args)

    call(context, builder, libawkward.ArrayBuilder_endlist, (proxyin.rawptr,))

    return context.get_dummy_value()


@numba.extending.lower_builtin(
    "append", ArrayBuilderType, ak._connect.numba.arrayview.RecordViewType
)
def lower_append_record(context, builder, sig, args):
    arraybuildertype, recordviewtype = sig.args
    arraybuilderval, recordviewval = args

    recordviewproxy = context.make_helper(builder, recordviewtype, recordviewval)

    arrayviewproxy = context.make_helper(
        builder, recordviewtype.arrayviewtype, recordviewproxy.arrayview
    )
    atval = ak._connect.numba.layout.castint(
        context, builder, numba.intp, numba.int64, recordviewproxy.at
    )

    sharedptr = ak._connect.numba.layout.getat(
        context, builder, arrayviewproxy.sharedptrs, arrayviewproxy.pos
    )

    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    call(
        context,
        builder,
        libawkward.ArrayBuilder_append_nowrap,
        (
            proxyin.rawptr,
            builder.inttoptr(sharedptr, context.get_value_type(numba.types.voidptr)),
            atval,
        ),
    )
    return context.get_dummy_value()


@numba.extending.lower_builtin("append", ArrayBuilderType, numba.types.Boolean)
def lower_append_bool(context, builder, sig, args):
    return lower_boolean(context, builder, sig, args)


@numba.extending.lower_builtin("append", ArrayBuilderType, numba.types.Integer)
def lower_append_int(context, builder, sig, args):
    return lower_integer(context, builder, sig, args)


@numba.extending.lower_builtin("append", ArrayBuilderType, numba.types.Float)
def lower_append_float(context, builder, sig, args):
    return lower_real(context, builder, sig, args)


@numba.extending.lower_builtin("append", ArrayBuilderType, numba.types.Complex)
def lower_append_complex(context, builder, sig, args):
    return lower_complex(context, builder, sig, args)


@numba.extending.lower_builtin("append", ArrayBuilderType, numba.types.NPDatetime)
def lower_append_datetime(context, builder, sig, args):
    return lower_datetime(context, builder, sig, args)


@numba.extending.lower_builtin("append", ArrayBuilderType, numba.types.NPTimedelta)
def lower_append_timedelta(context, builder, sig, args):
    return lower_timedelta(context, builder, sig, args)


@numba.extending.lower_builtin("append", ArrayBuilderType, numba.types.UnicodeType)
def lower_append_string(context, builder, sig, args):
    return lower_string(context, builder, sig, args)


@numba.extending.lower_builtin("append", ArrayBuilderType, numba.types.Optional)
def lower_append_optional(context, builder, sig, args):
    arraybuildertype, opttype = sig.args
    arraybuilderval, optval = args

    optproxy = context.make_helper(builder, opttype, optval)
    validbit = numba.core.cgutils.as_bool_bit(builder, optproxy.valid)

    with builder.if_else(validbit) as (is_valid, is_not_valid):
        with is_valid:
            if isinstance(opttype.type, numba.types.Boolean):
                lower_boolean(
                    context,
                    builder,
                    numba.types.none(arraybuildertype, opttype.type),
                    (arraybuilderval, optproxy.data),
                )
            elif isinstance(opttype.type, numba.types.Integer):
                lower_integer(
                    context,
                    builder,
                    numba.types.none(arraybuildertype, opttype.type),
                    (arraybuilderval, optproxy.data),
                )
            elif isinstance(opttype.type, numba.types.Float):
                lower_real(
                    context,
                    builder,
                    numba.types.none(arraybuildertype, opttype.type),
                    (arraybuilderval, optproxy.data),
                )
            elif isinstance(opttype.type, numba.types.Complex):
                lower_complex(
                    context,
                    builder,
                    numba.types.none(arraybuildertype, opttype.type),
                    (arraybuilderval, optproxy.data),
                )
            elif isinstance(opttype.type, numba.types.UnicodeType):
                lower_string(
                    context,
                    builder,
                    numba.types.none(arraybuildertype, opttype.type),
                    (arraybuilderval, optproxy.data),
                )
            else:
                raise AssertionError(repr(opttype.type))

        with is_not_valid:
            lower_null(
                context,
                builder,
                numba.types.none(
                    arraybuildertype,
                ),
                (arraybuilderval,),
            )

    return context.get_dummy_value()


@numba.extending.lower_builtin("append", ArrayBuilderType, numba.types.NoneType)
def lower_append_none(context, builder, sig, args):
    return lower_null(context, builder, sig.return_type(sig.args[0]), (args[0],))


@numba.extending.lower_builtin(
    "extend", ArrayBuilderType, ak._connect.numba.arrayview.ArrayViewType
)
def lower_extend_array(context, builder, sig, args):
    arraybuildertype, viewtype = sig.args
    arraybuilderval, viewval = args

    viewproxy = context.make_helper(builder, viewtype, viewval)

    sharedptr = ak._connect.numba.layout.getat(
        context, builder, viewproxy.sharedptrs, viewproxy.pos
    )

    proxyin = context.make_helper(builder, arraybuildertype, arraybuilderval)
    with numba.core.cgutils.for_range(builder, viewproxy.stop, viewproxy.start) as loop:
        atval = ak._connect.numba.layout.castint(
            context, builder, numba.intp, numba.int64, loop.index
        )
        call(
            context,
            builder,
            libawkward.ArrayBuilder_append_nowrap,
            (
                proxyin.rawptr,
                builder.inttoptr(
                    sharedptr, context.get_value_type(numba.types.voidptr)
                ),
                atval,
            ),
        )

    return context.get_dummy_value()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("from_arrow_schema",)
from awkward._dispatch import high_level_function
from awkward._nplikes.numpylike import NumpyMetadata

np = NumpyMetadata.instance()


@high_level_function
def from_arrow_schema(schema):
    """
    Args:
        schema (`pyarrow.Schema`): Apache Arrow schema to convert into an Awkward Form.

    Converts an Apache Arrow schema into an Awkward Form.

    See also #ak.to_arrow, #ak.to_arrow_table, #ak.from_arrow, #ak.to_parquet, #ak.from_parquet.
    """
    return _impl(schema)


def _impl(schema):
    import awkward._connect.pyarrow

    return awkward._connect.pyarrow.form_handle_arrow(schema, pass_empty_field=True)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

import re
import ast
import glob
import io
import subprocess
import pathlib
import os
import warnings

import sphinx.ext.napoleon

config = sphinx.ext.napoleon.Config(napoleon_use_param=True, napoleon_use_rtype=True)

reference_path = pathlib.Path("reference")
output_path = reference_path / "generated"
output_path.mkdir(exist_ok=True)

latest_commit = (
    subprocess.run(["git", "rev-parse", "HEAD"], stdout=subprocess.PIPE)
    .stdout.decode("utf-8")
    .strip()
)

toctree = []


def tostr(node):
    if isinstance(node, ast.NameConstant):
        return repr(node.value)

    elif isinstance(node, ast.Num):
        return repr(node.n)

    elif isinstance(node, ast.Str):
        return repr(node.s)

    elif isinstance(node, ast.Name):
        return node.id

    elif isinstance(node, ast.Index):
        return tostr(node.value)

    elif isinstance(node, ast.Attribute):
        return tostr(node.value) + "." + node.attr

    elif isinstance(node, ast.Subscript):
        return "{0}[{1}]".format(tostr(node.value), tostr(node.slice))

    elif isinstance(node, ast.Slice):
        start = "" if node.lower is None else tostr(node.lower)
        stop = "" if node.upper is None else tostr(node.upper)
        step = "" if node.step is None else tostr(node.step)
        if step == "":
            return "{0}:{1}".format(start, stop)
        else:
            return "{0}:{1}:{2}".format(start, stop, step)

    elif isinstance(node, ast.Call):
        return "{0}({1})".format(
            tostr(node.func), ", ".join(tostr(x) for x in node.args)
        )

    elif isinstance(node, ast.List):
        return "[{0}]".format(", ".join(tostr(x) for x in node.elts))

    elif isinstance(node, ast.Set):
        return "{{{0}}}".format(", ".join(tostr(x) for x in node.elts))

    elif isinstance(node, ast.Tuple):
        return "({0})".format(
            ", ".join(tostr(x) for x in node.elts)
            + ("," if len(node.elts) == 1 else "")
        )

    elif isinstance(node, ast.Dict):
        return "{{{0}}}".format(
            ", ".join(
                "{0}: {1}".format(tostr(x), tostr(y))
                for x, y in zip(node.keys, node.values)
            )
        )

    elif isinstance(node, ast.Lambda):
        return "lambda {0}: {1}".format(
            ", ".join(x.arg for x in node.args.args), tostr(node.body)
        )

    elif isinstance(node, ast.UnaryOp):
        return tostr.op[type(node.op)] + tostr(node.operand)

    elif isinstance(node, ast.BinOp):
        return tostr(node.left) + tostr.op[type(node.op)] + tostr(node.right)

    elif isinstance(node, ast.Compare):
        return tostr(node.left) + "".join(
            tostr.op[type(x)] + tostr(y) for x, y in zip(node.ops, node.comparators)
        )

    elif isinstance(node, ast.BoolOp):
        return tostr.op[type(node.op)].join(tostr(x) for x in node.values)

    else:
        raise Exception(ast.dump(node))


tostr.op = {
    ast.And: " and ",
    ast.Or: " or ",
    ast.Add: " + ",
    ast.Sub: " - ",
    ast.Mult: " * ",
    ast.MatMult: " @ ",
    ast.Div: " / ",
    ast.Mod: " % ",
    ast.Pow: "**",
    ast.LShift: " << ",
    ast.RShift: " >> ",
    ast.BitOr: " | ",
    ast.BitXor: " ^ ",
    ast.BitAnd: " & ",
    ast.FloorDiv: " // ",
    ast.Invert: "~",
    ast.Not: "not ",
    ast.UAdd: "+",
    ast.USub: "-",
    ast.Eq: " == ",
    ast.NotEq: " != ",
    ast.Lt: " < ",
    ast.LtE: " <= ",
    ast.Gt: " > ",
    ast.GtE: " >= ",
    ast.Is: " is ",
    ast.IsNot: " is not ",
    ast.In: " in ",
    ast.NotIn: " not in ",
}


def dosig(node):
    if node is None:
        return "self"
    else:
        non_keyword_args = node.args.posonlyargs + node.args.args
        argnames = [x.arg for x in non_keyword_args + node.args.kwonlyargs]
        defaults = [
            "=" + tostr(x)
            for x in node.args.defaults + node.args.kw_defaults
            if x is not None
        ]
        defaults = [""] * (len(argnames) - len(defaults)) + defaults
        rendered = [x + y for x, y in zip(argnames, defaults)]
        if node.args.vararg is not None:
            rendered.insert(len(non_keyword_args), f"*{node.args.vararg.arg}")
        elif node.args.kwonlyargs:
            rendered.insert(len(non_keyword_args), "*")
        return ", ".join(rendered)


def dodoc(docstring, qualname, names):
    out = docstring.replace("`", "``")
    out = re.sub(r"<<<([^>]*)>>>", r"`\1`_", out)
    out = re.sub(r"#(ak\.[A-Za-z0-9_\.]*[A-Za-z0-9_])", r":py:obj:`\1`", out)
    for x in names:
        out = out.replace("#" + x, ":py:meth:`{1} <{0}.{1}>`".format(qualname, x))
    out = re.sub(r"\[([^\]]*)\]\(([^\)]*)\)", r"`\1 <\2>`__", out)
    out = str(sphinx.ext.napoleon.GoogleDocstring(out, config))
    out = re.sub(
        r"([^\. \t].*\n[ \t]*)((\n    .*[^ \t].*)(\n    .*[^ \t].*|\n[ \t]*)*)",
        "\\1\n.. code-block:: python\n\n\\2",
        out,
    )
    out = re.sub(r"(\n:param|^:param)", "\n    :param", out)
    out = re.sub(r"(\n:type|^:type)", "\n    :type", out)
    out = re.sub(r"(\n:returns|^:returns)", "\n    :returns", out)
    out = re.sub(r"(\n:raises|^:raises)", "\n    :raises", out)
    return out


def make_anchor(name):
    name = name.lower()
    parts = name.split(".")
    anchor = "-".join(parts)
    return f"\n\n.. _{anchor}:\n\n"


def doclass(link, linelink, shortname, name, astcls):
    if name.startswith("_"):
        return

    qualname = shortname + "." + name

    init, rest, names = None, [], []
    for node in astcls.body:
        if isinstance(node, ast.FunctionDef):
            if node.name == "__init__":
                init = node
            else:
                rest.append(node)
                names.append(node.name)
        elif (
            isinstance(node, ast.Assign)
            and len(node.targets) == 1
            and isinstance(node.targets[0], ast.Name)
        ):
            rest.append(node)
            names.append(node.targets[0].id)

    outfile = io.StringIO()
    outfile.write(qualname + "\n" + "-" * len(qualname) + "\n\n")
    outfile.write(f".. py:module: {qualname}\n\n")
    outfile.write("Defined in {0}{1}.\n\n".format(link, linelink))
    outfile.write(".. py:class:: {0}({1})\n\n".format(qualname, dosig(init)))

    docstring = ast.get_docstring(astcls)
    if docstring is not None:
        outfile.write(dodoc(docstring, qualname, names) + "\n\n")

    for node in rest:
        if isinstance(node, ast.Assign):
            attrtext = "{0}.{1}".format(qualname, node.targets[0].id)
            outfile.write(make_anchor(attrtext))
            outfile.write(".. py:attribute:: " + attrtext + "\n")
            outfile.write("    :value: {0}\n\n".format(tostr(node.value)))
            docstring = None

        elif any(
            isinstance(x, ast.Name) and x.id == "property" for x in node.decorator_list
        ):
            attrtext = "{0}.{1}".format(qualname, node.name)
            outfile.write(make_anchor(attrtext))
            outfile.write(".. py:attribute:: " + attrtext + "\n\n")
            docstring = ast.get_docstring(node)

        elif any(
            isinstance(x, ast.Attribute) and x.attr == "setter"
            for x in node.decorator_list
        ):
            docstring = None

        else:
            methodname = "{0}.{1}".format(qualname, node.name)
            methodtext = "{0}({1})".format(methodname, dosig(node))
            outfile.write(make_anchor(methodname))
            outfile.write(".. py:method:: " + methodtext + "\n\n")
            docstring = ast.get_docstring(node)

        if docstring is not None:
            outfile.write(dodoc(docstring, qualname, names) + "\n\n")

    toctree.append(os.path.join("generated", qualname + ".rst"))
    out = outfile.getvalue()
    entry_path = reference_path / toctree[-1]
    if not (entry_path.exists() and entry_path.read_text() == out):
        print("writing", toctree[-1])
        entry_path.write_text(out)


def dofunction(link, linelink, shortname, name, astfcn):
    if name.startswith("_"):
        return

    qualname = shortname + "." + name

    outfile = io.StringIO()
    outfile.write(qualname + "\n" + "-" * len(qualname) + "\n\n")
    outfile.write(f".. py:module: {qualname}\n\n")
    outfile.write("Defined in {0}{1}.\n\n".format(link, linelink))

    functiontext = "{0}({1})".format(qualname, dosig(astfcn))
    outfile.write(".. py:function:: " + functiontext + "\n\n")

    docstring = ast.get_docstring(astfcn)
    if docstring is not None:
        outfile.write(dodoc(docstring, qualname, []) + "\n\n")

    out = outfile.getvalue()

    toctree.append(os.path.join("generated", qualname + ".rst"))

    entry_path = reference_path / toctree[-1]
    if not (entry_path.exists() and entry_path.read_text() == out):
        print("writing", toctree[-1])
        entry_path.write_text(out)


for filename in glob.glob("../src/awkward/**/*.py", recursive=True):

    modulename = (
        filename.replace("../src/", "")
        .replace("/__init__.py", "")
        .replace(".py", "")
        .replace("/", ".")
    )

    shortname = (
        modulename.replace("awkward.", "ak.")
        .replace(".highlevel", "")
        .replace(".behaviors.mixins", "")
        .replace(".behaviors.categorical", "")
        .replace(".behaviors.string", "")
    )
    shortname = re.sub(r"\.operations\.ak_\w+", "", shortname)
    shortname = re.sub(r"\.(contents|types|forms)\.\w+", r".\1", shortname)

    if (
        modulename.startswith("awkward._")
        or modulename == "awkward.nplikes"
        or modulename == "awkward.types._awkward_datashape_parser"
    ):
        continue  # don't show awkward._*, including _v2

    link = "`{0} <https://github.com/scikit-hep/awkward-1.0/blob/" "{1}/{2}>`__".format(
        modulename, latest_commit, filename.replace("../", "")
    )

    module = ast.parse(open(filename).read())

    for toplevel in module.body:
        if hasattr(toplevel, "lineno"):
            linelink = (
                " on `line {0} <https://github.com/scikit-hep/awkward-1.0/blob/"
                "{1}/{2}#L{0}>`__".format(
                    toplevel.lineno, latest_commit, filename.replace("../", "")
                )
            )
        else:
            lineline = ""
        if isinstance(toplevel, ast.ClassDef):
            doclass(link, linelink, shortname, toplevel.name, toplevel)
        if isinstance(toplevel, ast.FunctionDef):
            dofunction(link, linelink, shortname, toplevel.name, toplevel)


def test_signature_pos_or_kw():
    mod = ast.parse(
        """
def func(x, z):
    ...
"""
    )
    node = mod.body[0]
    assert dosig(node) == "x, z"


def test_signature_kwarg():
    mod = ast.parse(
        """
def func(x, z=None):
    ...
"""
    )
    node = mod.body[0]
    assert dosig(node) == "x, z=None"


def test_signature_vararg():
    mod = ast.parse(
        """
def func(x, *y, z=None):
    ...
"""
    )
    node = mod.body[0]
    assert dosig(node) == "x, *y, z=None"


def test_signature_kwonly():
    mod = ast.parse(
        """
def func(x, *, y, z=None):
    ...
"""
    )
    node = mod.body[0]
    assert dosig(node) == "x, *, y, z=None"


test_signature_pos_or_kw()
test_signature_kwarg()
test_signature_vararg()
test_signature_kwonly()


toctree_path = reference_path / "toctree.txt"
toctree_contents = toctree_path.read_text()

for p in toctree:
    if p.removesuffix(".rst") not in toctree_contents:
        warnings.warn(f"Couldn't find {p} in toctree contents")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test_sort():
    data = ak.Array([[7, 5, 7], [], [2], [8, 2]])
    assert to_list(ak.operations.sort(data)) == [
        [5, 7, 7],
        [],
        [2],
        [2, 8],
    ]


def test_argsort():
    data = ak.Array([[7, 5, 7], [], [2], [8, 2]])
    index = ak.operations.argsort(data)
    assert to_list(data[index]) == [[5, 7, 7], [], [2], [2, 8]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("with_parameter",)
import awkward as ak
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata

np = NumpyMetadata.instance()


@high_level_function
def with_parameter(array, parameter, value, *, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        parameter (str): Name of the parameter to set on that array.
        value (JSON): Value of the parameter to set on that array.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    This function returns a new array with a parameter set on the outermost
    node of its #ak.Array.layout.

    Note that a "new array" is a lightweight shallow copy, not a duplication
    of large data buffers.

    You can also remove a single parameter with this function, since setting
    a parameter to None is equivalent to removing it.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, parameter, value, highlevel, behavior)


def _impl(array, parameter, value, highlevel, behavior):
    behavior = behavior_of(array, behavior=behavior)
    layout = ak.operations.to_layout(array, allow_record=True, allow_other=False)

    out = layout.with_parameter(parameter, value)

    return wrap_layout(out, behavior_of(array, behavior=behavior), highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_trailing_sublist():
    layout = ak.contents.ListOffsetArray(
        ak.index.Index(np.array([0, 6, 9, 9], dtype=np.int64)),
        ak.contents.NumpyArray(np.array([1, 1, 2, 2, 2, 3, 4, 4, 5])),
    )
    assert ak.run_lengths(layout).to_list() == [[2, 3, 1], [2, 1], []]


def test_leading_sublist():
    layout = ak.contents.ListOffsetArray(
        ak.index.Index(np.array([0, 0, 6, 9], dtype=np.int64)),
        ak.contents.NumpyArray(np.array([1, 1, 2, 2, 2, 3, 4, 4, 5])),
    )
    assert ak.run_lengths(layout).to_list() == [[], [2, 3, 1], [2, 1]]


def test_leading_trailing_sublist():
    layout = ak.contents.ListOffsetArray(
        ak.index.Index(np.array([0, 0, 6, 9, 9], dtype=np.int64)),
        ak.contents.NumpyArray(np.array([1, 1, 2, 2, 2, 3, 4, 4, 5])),
    )
    assert ak.run_lengths(layout).to_list() == [[], [2, 3, 1], [2, 1], []]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


def test_simple():
    left = ak.Array([1, 2, 3])
    right = ak.Array([["lambda", "sigma", "eta", "phi"], ["delta"]])
    pair = ak.cartesian([left, right], axis=0)
    assert pair.ndim == 1
    assert pair.tolist() == [
        (1, ["lambda", "sigma", "eta", "phi"]),
        (1, ["delta"]),
        (2, ["lambda", "sigma", "eta", "phi"]),
        (2, ["delta"]),
        (3, ["lambda", "sigma", "eta", "phi"]),
        (3, ["delta"]),
    ]


def test_out_of_bounds():
    left = ak.Array([1, 2, 3])
    right = ak.Array([["lambda", "sigma", "eta", "phi"], ["delta"]])
    with pytest.raises(np.AxisError):
        ak.cartesian([left, right], axis=2)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401

import awkward as ak


def test_indexed():
    x = ak.to_layout([{"a": 1, "b": 2}])
    y = ak.contents.IndexedArray(
        ak.index.Index64([1]), ak.to_layout([{"c": 13, "b": 15}, {"c": 3, "b": 5}])
    )

    z = ak.concatenate((x, y))

    assert z.tolist() == [
        {"a": 1, "b": 2},
        {"c": 3, "b": 5},
    ]
    assert z.type == ak.types.ArrayType(
        ak.types.UnionType(
            [
                ak.types.RecordType(
                    [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                    ["a", "b"],
                ),
                ak.types.RecordType(
                    [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                    ["c", "b"],
                ),
            ]
        ),
        2,
    )

    w = ak.merge_union_of_records(z)
    assert w.type == ak.types.ArrayType(
        ak.types.RecordType(
            [
                ak.types.OptionType(ak.types.NumpyType("int64")),
                ak.types.NumpyType("int64"),
                ak.types.OptionType(ak.types.NumpyType("int64")),
            ],
            ["a", "b", "c"],
        ),
        2,
    )
    assert w.tolist() == [{"a": 1, "b": 2, "c": None}, {"a": None, "b": 5, "c": 3}]


def test_option():
    x = ak.to_layout([{"a": 1, "b": 2}])
    y = ak.to_layout([{"c": 3, "b": 5}, None])

    z = ak.concatenate((x, y))

    assert z.tolist() == [{"a": 1, "b": 2}, {"c": 3, "b": 5}, None]
    assert z.type == ak.types.ArrayType(
        ak.types.UnionType(
            [
                ak.types.OptionType(
                    ak.types.RecordType(
                        [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                        ["a", "b"],
                    )
                ),
                ak.types.OptionType(
                    ak.types.RecordType(
                        [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                        ["c", "b"],
                    )
                ),
            ]
        ),
        3,
    )

    w = ak.merge_union_of_records(z)
    assert w.type == ak.types.ArrayType(
        ak.types.OptionType(
            ak.types.RecordType(
                [
                    ak.types.OptionType(ak.types.NumpyType("int64")),
                    ak.types.NumpyType("int64"),
                    ak.types.OptionType(ak.types.NumpyType("int64")),
                ],
                ["a", "b", "c"],
            )
        ),
        3,
    )
    assert w.tolist() == [
        {"a": 1, "b": 2, "c": None},
        {"a": None, "b": 5, "c": 3},
        None,
    ]


def test_option_unmasked():
    x = ak.to_layout([{"a": 1, "b": 2}])
    y = ak.contents.UnmaskedArray(ak.to_layout([{"c": 3, "b": 5}]))

    z = ak.concatenate((x, y))

    assert z.tolist() == [{"a": 1, "b": 2}, {"c": 3, "b": 5}]
    assert z.type == ak.types.ArrayType(
        ak.types.UnionType(
            [
                ak.types.OptionType(
                    ak.types.RecordType(
                        [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                        ["a", "b"],
                    )
                ),
                ak.types.OptionType(
                    ak.types.RecordType(
                        [ak.types.NumpyType("int64"), ak.types.NumpyType("int64")],
                        ["c", "b"],
                    )
                ),
            ]
        ),
        2,
    )

    w = ak.merge_union_of_records(z)
    assert w.type == ak.types.ArrayType(
        ak.types.OptionType(
            ak.types.RecordType(
                [
                    ak.types.OptionType(ak.types.NumpyType("int64")),
                    ak.types.NumpyType("int64"),
                    ak.types.OptionType(ak.types.NumpyType("int64")),
                ],
                ["a", "b", "c"],
            )
        ),
        2,
    )
    assert w.tolist() == [{"a": 1, "b": 2, "c": None}, {"a": None, "b": 5, "c": 3}]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


def test_types_with_parameters():
    t = ak.types.UnknownType()
    assert t.parameters == {}

    with pytest.raises(TypeError):
        ak.types.UnknownType(parameters={"latitude": ["val", "ue"]})
    t = ak.types.UnknownType()
    assert t.parameters == {}

    t = ak.types.NumpyType("int32", parameters={"latitude": ["val", "ue"]})
    assert t.parameters == {"latitude": ["val", "ue"]}
    t = ak.types.NumpyType("float64", parameters={"latitude": ["val", "ue"]})
    assert t.parameters == {"latitude": ["val", "ue"]}
    t = ak.types.ArrayType(
        ak.types.NumpyType("int32", parameters={"latitude": ["val", "ue"]}), 100
    )
    assert t.content.parameters == {"latitude": ["val", "ue"]}
    t = ak.types.ListType(
        ak.types.NumpyType("int32"), parameters={"latitude": ["val", "ue"]}
    )
    assert t.parameters == {"latitude": ["val", "ue"]}
    t = ak.types.RegularType(
        ak.types.NumpyType("int32"), 5, parameters={"latitude": ["val", "ue"]}
    )
    assert t.parameters == {"latitude": ["val", "ue"]}
    t = ak.types.OptionType(
        ak.types.NumpyType("int32"), parameters={"latitude": ["val", "ue"]}
    )
    assert t.parameters == {"latitude": ["val", "ue"]}
    t = ak.types.UnionType(
        (ak.types.NumpyType("int32"), ak.types.NumpyType("float64")),
        parameters={"latitude": ["val", "ue"]},
    )
    assert t.parameters == {"latitude": ["val", "ue"]}
    t = ak.types.RecordType(
        [
            ak.types.NumpyType("int32"),
            ak.types.NumpyType("float64"),
        ],
        fields=["one", "two"],
        parameters={"latitude": ["val", "ue"]},
    )
    assert t.parameters == {"latitude": ["val", "ue"]}


def test_dress():
    class Dummy(ak.highlevel.Array):
        def __str__(self):
            return f"<Dummy {super().__str__()}>"

    ns = {"Dummy": Dummy}

    x = ak.contents.RegularArray(
        ak.contents.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6, 6])),
        size=3,
        parameters={"__list__": "Dummy"},
    )
    a = ak.highlevel.Array(x, behavior=ns, check_valid=True)
    assert str(a) == "<Dummy [[1.1, 2.2, 3.3], [4.4, 5.5, 6]]>"

    x2 = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 2, 2, 2], dtype=np.int64)),
        x,
    )
    a2 = ak.highlevel.Array(x2, behavior=ns, check_valid=True)
    # columns limit changed from 40 to 80 in v2
    assert (
        repr(a2)
        == "<Array [<Dummy [[1.1, 2.2, 3.3], [4.4, 5.5, 6]]>, <Dummy []>, <Dummy []>] type='...'>"
    )
    assert str(a2[0]) == "<Dummy [[1.1, 2.2, 3.3], [4.4, 5.5, 6]]>"
    assert str(a2[1]) == "<Dummy []>"
    assert str(a2[2]) == "<Dummy []>"


def test_record_name():
    builder = ak.highlevel.ArrayBuilder()

    builder.begin_record("Dummy")
    builder.field("one")
    builder.integer(1)
    builder.field("two")
    builder.real(1.1)
    builder.end_record()

    builder.begin_record("Dummy")
    builder.field("two")
    builder.real(2.2)
    builder.field("one")
    builder.integer(2)
    builder.end_record()

    a = builder.snapshot()

    assert str(a.layout.form.type) == "Dummy[one: int64, two: float64]"
    assert a.layout.form.type.parameters == {"__record__": "Dummy"}


def test_builder_string():
    builder = ak.ArrayBuilder()
    builder.bytestring(b"one")
    builder.bytestring(b"two")
    builder.bytestring(b"three")

    a1 = builder.snapshot()
    assert str(a1) == "[b'one', b'two', b'three']"

    builder = ak.highlevel.ArrayBuilder()
    builder.bytestring(b"one")
    builder.bytestring(b"two")
    builder.bytestring(b"three")

    a = builder.snapshot()

    assert str(a) == "[b'one', b'two', b'three']"
    assert to_list(a) == [b"one", b"two", b"three"]
    assert (
        ak.operations.to_json(a, convert_bytes=bytes.decode) == '["one","two","three"]'
    )
    assert repr(a) == "<Array [b'one', b'two', b'three'] type='3 * bytes'>"
    assert str(ak.operations.type(a)) == "3 * bytes"

    builder = ak.highlevel.ArrayBuilder()

    builder.string("one")
    builder.string("two")
    builder.string("three")

    a = builder.snapshot()
    assert str(a) == "['one', 'two', 'three']"
    assert to_list(a) == ["one", "two", "three"]
    assert ak.operations.to_json(a) == '["one","two","three"]'
    assert repr(a) == "<Array ['one', 'two', 'three'] type='3 * string'>"
    assert str(ak.operations.type(a)) == "3 * string"

    builder = ak.highlevel.ArrayBuilder()

    builder.begin_list()
    builder.string("one")
    builder.string("two")
    builder.string("three")
    builder.end_list()

    builder.begin_list()
    builder.end_list()

    builder.begin_list()
    builder.string("four")
    builder.string("five")
    builder.end_list()

    a = builder.snapshot()
    assert str(a) == "[['one', 'two', 'three'], [], ['four', 'five']]"
    assert to_list(a) == [["one", "two", "three"], [], ["four", "five"]]
    assert ak.operations.to_json(a) == '[["one","two","three"],[],["four","five"]]'
    assert str(ak.operations.type(a)) == "3 * var * string"


def test_fromiter_fromjson():
    assert to_list(ak.from_iter(["one", "two", "three"])) == ["one", "two", "three"]
    assert to_list(ak.from_iter([["one", "two", "three"], [], ["four", "five"]])) == [
        ["one", "two", "three"],
        [],
        ["four", "five"],
    ]


def test_fromjson():
    assert to_list(ak.operations.from_json('["one", "two", "three"]')) == [
        "one",
        "two",
        "three",
    ]
    assert to_list(
        ak.operations.from_json('[["one", "two", "three"], [], ["four", "five"]]')
    ) == [["one", "two", "three"], [], ["four", "five"]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import glob
import math
import os

import numpy

try:
    import cupy

    error_message = None

except ModuleNotFoundError:
    cupy = None
    error_message = """to use {0}, you must install cupy:

    pip install cupy

or

    conda install -c conda-forge cupy
"""

cuda_streamptr_to_contexts = {}
kernel_errors = {}
kernel = None

ERROR_BITS = 8
NO_ERROR = numpy.iinfo(numpy.uint64).max


dtype_to_ctype = {
    numpy.bool_: "bool",
    numpy.int8: "int8_t",
    numpy.uint8: "uint8_t",
    numpy.int16: "int16_t",
    numpy.uint16: "uint16_t",
    numpy.int32: "int32_t",
    numpy.uint32: "uint32_t",
    numpy.int64: "int64_t",
    numpy.uint64: "uint64_t",
    numpy.float32: "float",
    numpy.float64: "double",
}


def fetch_specialization(keys):
    specialized_name = keys[0].replace("'", "") + "<"

    keys = keys[1:]
    for key in keys[:-1]:
        if dtype_to_ctype.get(key) is None:
            dtype_string = repr(key)
            specialized_name = specialized_name + dtype_string.split("'")[1] + "_t, "
        else:
            specialized_name = specialized_name + dtype_to_ctype[key] + ", "

    if dtype_to_ctype.get(keys[-1]) is None:
        dtype_string = repr(keys[-1])
        specialized_name = specialized_name + dtype_string.split("'")[1] + "_t>"
    else:
        specialized_name = specialized_name + dtype_to_ctype[keys[-1]] + ">"

    return specialized_name


def fetch_template_specializations(kernel_dict):
    # These kernels consist of multiple kernels don't have templated specializations of the same name
    kernel_exclusions = [
        "awkward_ByteMaskedArray_getitem_nextcarry",
        "awkward_ByteMaskedArray_getitem_nextcarry_outindex",
        "awkward_ByteMaskedArray_reduce_next_64",
        "awkward_ByteMaskedArray_reduce_next_nonlocal_nextshifts_64",
        "awkward_Content_getitem_next_missing_jagged_getmaskstartstop",
        "awkward_IndexedArray_flatten_nextcarry",
        "awkward_IndexedArray_getitem_nextcarry",
        "awkward_IndexedArray_getitem_nextcarry_outindex",
        "awkward_IndexedArray_getitem_nextcarry_outindex_mask",
        "awkward_ListArray_compact_offsets",
        "awkward_IndexedArray_reduce_next_64",
        "awkward_IndexedArray_reduce_next_nonlocal_nextshifts_64",
        "awkward_IndexedArray_reduce_next_nonlocal_nextshifts_fromshifts_64",
        "awkward_IndexedOptionArray_rpad_and_clip_mask_axis1",
        "awkward_MaskedArray_getitem_next_jagged_project",
        "awkward_NumpyArray_getitem_boolean_nonzero",
        "awkward_UnionArray_project",
        "awkward_reduce_count_64",
        "awkward_reduce_sum",
        "awkward_reduce_sum_int32_bool_64",
        "awkward_reduce_sum_int64_bool_64",
        "awkward_reduce_sum_bool",
        "awkward_reduce_prod_bool",
        "awkward_reduce_argmax",
        "awkward_reduce_argmax_bool_64",
        "awkward_reduce_argmin",
        "awkward_reduce_argmin_bool_64",
        "awkward_reduce_countnonzero",
        "awkward_reduce_max",
        "awkward_reduce_min",
    ]
    template_specializations = []
    import re

    for keys, value in kernel_dict.items():
        pattern = re.compile("_[a-z]$")
        if keys[0] not in kernel_exclusions:
            if value is None:
                if pattern.search(keys[0]):
                    template_specializations.append(fetch_specialization(list(keys)))
            else:
                template_specializations.append(fetch_specialization(list(keys)))

    return template_specializations


def populate_kernel_errors(kernel_name, cu_file):
    import re

    pattern_errstring = "// message:"

    result_errstring = [_.start() for _ in re.finditer(pattern_errstring, cu_file)]

    err_strings = []
    for index in result_errstring:
        error = cu_file[index:]
        error = error[error.find('"') : error.find("\n")]
        error = error.replace('"', "")
        err_strings.append(error)

    kernel_errors[kernel_name] = err_strings


class Invocation:
    def __init__(self, name, error_context):
        self._name = name
        self._error_context = error_context

    @property
    def name(self):
        return self._name

    @property
    def error_context(self):
        return self._error_context


def import_cupy(name="Awkward Arrays with CUDA"):
    if cupy is None:
        raise ModuleNotFoundError(error_message.format(name))

    return cupy


def initialize_cuda_kernels(cupy):
    if cupy is not None:
        global kernel

        if kernel is None:
            import awkward._connect.cuda._kernel_signatures

            cuda_src = f"#define ERROR_BITS {ERROR_BITS}\n#define NO_ERROR {NO_ERROR}"

            cuda_kernels_path = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "cuda_kernels"
            )

            with open(
                os.path.join(cuda_kernels_path, "cuda_common.cu"),
                encoding="utf-8",
            ) as header_file:
                cuda_src = cuda_src + "\n" + header_file.read()
            for filename in glob.glob(os.path.join(cuda_kernels_path, "awkward_*.cu")):
                with open(filename, encoding="utf-8") as cu_file:
                    cu_code = cu_file.read()
                    populate_kernel_errors(
                        filename[filename.find("awkward_") : filename.find(".cu")],
                        cu_code,
                    )
                    cuda_src = cuda_src + "\n" + cu_code

            # Pass an empty Raw Module to fetch all template specializations
            template_specializations = fetch_template_specializations(
                awkward._connect.cuda._kernel_signatures.by_signature(None)
            )
            cuda_kernel_templates = cupy.RawModule(
                code=cuda_src,
                options=("--std=c++11",),
                name_expressions=template_specializations,
            )
            kernel = awkward._connect.cuda._kernel_signatures.by_signature(
                cuda_kernel_templates
            )

        return kernel
    else:
        raise ModuleNotFoundError(error_message.format("Awkward Arrays with CUDA"))


def synchronize_cuda(stream=None):
    cupy = import_cupy("Awkward Arrays with CUDA")

    if stream is None:
        stream = cupy.cuda.get_current_stream()

    stream.synchronize()

    invocation_index = cuda_streamptr_to_contexts[stream.ptr][0].get().tolist()
    contexts = cuda_streamptr_to_contexts[stream.ptr][1]

    if invocation_index != NO_ERROR:
        invoked_kernel = contexts[int(invocation_index // math.pow(2, ERROR_BITS))]
        cuda_streamptr_to_contexts[stream.ptr] = (
            cupy.array(NO_ERROR),
            [],
        )
        raise invoked_kernel.error_context.decorate_exception(
            ValueError(
                f"{kernel_errors[invoked_kernel.name][int(invocation_index % math.pow(2, ERROR_BITS))]} in compiled CUDA code ({invoked_kernel.name})"
            )
        )
